# -*- coding: utf-8 -*-
"""Class for tests of pysiaalarm."""
import logging
import threading
import asyncio
import pytest
from dataclasses import asdict

from pytest_cases import parametrize_with_cases, fixture
from unittest.mock import patch

from pysiaalarm import (
    SIAAccount,
    SIAClient,
    SIAEvent,
)
from pysiaalarm.aio import SIAClient as SIAClientA
from pysiaalarm.aio.client import SIAClientTCP
from pysiaalarm.event import NAKEvent, BaseEvent
from pysiaalarm.const import COUNTER_USER_CODE, COUNTER_VALID, COUNTER_EVENTS
from pysiaalarm.errors import NoAccountError

from tests.test_alarm import send_messages
from tests.test_utils import ACCOUNT, KEY, HOST, create_test_line
from tests.test_sia_package_cases import *  # pylint: disable=W0614


logging.basicConfig(level=logging.INFO)
_LOGGER = logging.getLogger(__name__)


@pytest.fixture
def account():
    """Return default account."""
    return SIAAccount(ACCOUNT, KEY)


@pytest.fixture
def account_list(account):
    """Return default account."""
    return [account]


@fixture(unpack_into="client, config, good, sync")
@pytest.mark.asyncio
@parametrize_with_cases("good", prefix="handler_")
@parametrize_with_cases("key", prefix="encrypted_")
@parametrize_with_cases("sync", prefix="sync_")
@parametrize_with_cases("protocol", prefix="proto_")
def create_client(
    unused_tcp_port_factory,
    event_handler,
    async_event_handler,
    bad_handler,
    async_bad_handler,
    key,
    good,
    protocol,
    sync,
):
    """Fixture to create a SIA Client."""
    config = {
        "host": HOST,
        "port": unused_tcp_port_factory(),
        "account_id": ACCOUNT,
        "key": key,
        "protocol": protocol,
    }

    if sync:
        client = SIAClient(
            host=config["host"],
            port=config["port"],
            accounts=[SIAAccount(config["account_id"], config["key"])],
            function=event_handler if good else bad_handler,
            protocol=config["protocol"],
        )
    else:
        client = SIAClientA(
            host=config["host"],
            port=config["port"],
            accounts=[SIAAccount(config["account_id"], config["key"])],
            function=async_event_handler if good else async_bad_handler,
            protocol=config["protocol"],
        )
    return (client, config, good, sync)


def get_func(type="sync"):
    """Get a test func."""

    async def async_func(event: SIAEvent):
        """Pass for testing."""
        pass

    def func(event: SIAEvent):
        """Pass for testing."""
        pass

    if type == "both":
        return (func, async_func)
    elif type == "aio":
        return async_func
    else:
        return func


@fixture(
    unpack_into="events, event_handler, async_event_handler, bad_handler, async_bad_handler"
)
def get_event_func():
    """Return the handler function."""
    events = []

    async def async_event_handler(event: SIAEvent):
        events.append(event)

    def event_handler(event: SIAEvent):
        events.append(event)

    def bad_handler(_):
        raise ValueError("test error in user func")

    async def async_bad_handler(_):
        raise ValueError("test error in user func")

    return (events, event_handler, async_event_handler, bad_handler, async_bad_handler)


class testSIA(object):
    """Class for pysiaalarm tests."""

    pytestmark = pytest.mark.asyncio

    @parametrize_with_cases("cl, dic, clear_account", cases=ToFromDict)
    def test_to_from_dict(self, cl, dic, clear_account):
        """Test the to and from dict methods."""
        _LOGGER.warning("Dict before: %s", dic)
        instance = cl.from_dict(dic)
        _LOGGER.warning("Class instance %s", instance)
        if cl == SIAAccount:
            assert instance.account_id == dic["account_id"]
            assert instance.key == dic["key"]
            dic2 = instance.to_dict()
        else:
            assert instance.account == dic["account"]
            assert instance.code == dic["code"]
            dic2 = instance.to_dict()

        _LOGGER.warning("Dict after: %s", dic2)
        for key, value in dic.items():
            if key == "sia_account":
                if dic[key]:
                    assert instance.sia_account is not None
            elif key == "message_type":
                assert dic2[key].value == value
            elif key == "timestamp":
                assert dic2[key] == value.isoformat()
            else:
                assert dic2[key] == value

    @parametrize_with_cases(
        "line, account_id, code_type, code, error_type, extended_data_flag, encrypted_flag",
        cases=EventParsing,
    )
    def test_event_parsing(
        self,
        line,
        account_id,
        code_type,
        code,
        error_type,
        extended_data_flag,
        encrypted_flag,
    ):
        """Test event parsing methods."""
        catch = error_type if error_type is not None else Exception
        try:
            if extended_data_flag:
                event = SIAEvent.from_line(
                    line,
                    {
                        account_id: SIAAccount(
                            account_id, KEY, allowed_timeband=None  # type: ignore
                        )
                    },
                )
            else:
                event = SIAEvent.from_line(line)
            assert event.code == code
            if not event.valid_message:
                _LOGGER.warning("Event full message: %s", event.full_message)
                _LOGGER.warning("Event sent crc: %s", event.msg_crc)
                _LOGGER.warning("Event calc crc: %s", event.calc_crc)
            # assert event.valid_message
            if code:
                assert event.sia_code
                assert event.sia_code.type == code_type
            assert event.account == account_id
            if extended_data_flag:
                _LOGGER.warning("CRC: %s", event.calc_crc)
                assert event.extended_data is not None
                if event.extended_data[0].identifier == "K":
                    _LOGGER.debug("RSP Response: %s", event.create_response())
                    assert event.response == ResponseType.RSP
                else:
                    assert event.response == ResponseType.ACK
        except NoAccountError:
            assert encrypted_flag
        except AssertionError as e:
            raise e
        except catch as e:
            if error_type is None:
                _LOGGER.debug("Error thrown: %s", e)
                assert False
            if isinstance(e, error_type):
                assert True

    @parametrize_with_cases(
        "line, account_id, code_type, code, error_type, extended_data_flag, encrypted_flag",
        cases=EventParsing,
    )
    def test_event_serialization(
        self,
        line,
        account_id,
        code_type,
        code,
        error_type,
        extended_data_flag,
        encrypted_flag,
    ):
        """Test event parsing methods."""
        if error_type is not None:
            pytest.skip("Wrong messages will not serialize.")
            return
        if encrypted_flag is True:
            pytest.skip("Encrypted messages without account will not serialize.")
            return
        event = BaseEvent.from_line(line)
        event_type = event.__class__
        event_dict = event.to_dict()
        event2 = event_type.from_dict(event_dict)
        assert event == event2

    @parametrize_with_cases("key, account_id, error_type", cases=AccountSetup)
    def test_sia_account_setup(
        self, unused_tcp_port_factory, key, account_id, error_type
    ):
        """Test sia client behaviour."""
        try:
            SIAAccount.validate_account(account_id, key)
            SIAClient(
                host="",
                port=unused_tcp_port_factory(),
                accounts=[SIAAccount(account_id=account_id, key=key)],
                function=get_func("sync"),
            )
            assert True if error_type is None else False
        except Exception as exp:
            assert isinstance(exp, error_type)

    @parametrize_with_cases("key, account_id, error_type", cases=AccountSetup)
    def test_sia_account_validation(self, key, account_id, error_type):
        """Test sia account validation."""
        try:
            SIAAccount.validate_account(account_id, key)
            assert True if error_type is None else False
        except Exception as exp:
            assert isinstance(exp, error_type)

    @parametrize_with_cases("fault", prefix="fault_")
    async def test_clients(
        self,
        #event_loop,
        unused_tcp_port_factory,
        events,
        client,
        config,
        good,
        sync,
        fault,
    ):
        """Test the clients."""
        if sync:
            client.start()
        else:
            await client.async_start()
        await asyncio.sleep(0.01)

        t = threading.Thread(
            target=send_messages,
            name="send_messages",
            args=(config, {fault: True}),
        )
        t.daemon = True
        t.start()  # stops after the event has been sent.
        await asyncio.sleep(0.1)

        if sync:
            client.stop()
        else:
            await client.async_stop()

        _LOGGER.warning("Registered events: %s", client.counts)

        if fault is not None:
            assert client.counts.get(fault) == 1
        else:
            assert client.counts.get(COUNTER_VALID) == 1
        if good and not fault:
            assert len(events) == 1
            assert client.counts.get(COUNTER_EVENTS) == 1
        if not good and not fault:
            assert client.counts.get(COUNTER_USER_CODE) == 1
        if not good or fault:
            assert len(events) == 0

    @parametrize_with_cases("msg_type", prefix="msg_")
    @parametrize_with_cases(
        "code, alter_key, wrong_event, response", cases=ParseAndCheckEvent
    )
    @parametrize_with_cases("key", prefix="encrypted_")
    def test_parse_and_check(
        self,
        unused_tcp_port_factory,
        event_handler,
        key,
        code,
        msg_type,
        alter_key,
        wrong_event,
        response,
    ):
        """Test the server parsing."""
        if msg_type in ("ADM-CID", "NULL") and code == "ZX":
            pytest.skip(
                "Unknown code is not usefull to test in ADM-CID and NULL messages."
            )
            return
        if alter_key and key is None:
            pytest.skip("Alter key is not usefull to test when not encrypted.")
            return

        siac = SIAClient(
            HOST, unused_tcp_port_factory(), [SIAAccount(ACCOUNT, key)], event_handler
        )
        line = create_test_line(
            account=ACCOUNT,
            key=key,
            code=code,
            msg_type=msg_type,
            wrong_event=wrong_event,
            alter_key=alter_key,
        )

        _LOGGER.warning("Line to parse: %s", line)
        event = siac.sia_server.parse_and_check_event(line.encode())
        _LOGGER.warning("Event received: %s", event)

        assert event.response == response
        if wrong_event or alter_key:
            assert isinstance(event, NAKEvent)
        else:
            assert event.account == ACCOUNT
            assert event.code == "RP" if msg_type == "NULL" else code

    def test_accounts_get_set(self, account_list, unused_tcp_port):
        """Test the account getting and setting."""
        siac = SIAClient(
            host="",
            port=unused_tcp_port,
            accounts=account_list,
            function=(lambda ev: print(ev)),
        )
        assert siac.accounts == account_list
        acc_list2 = [
            SIAAccount(account_id=ACCOUNT, key=KEY, allowed_timeband=(None, None)),
            SIAAccount(account_id="1112", key=KEY, allowed_timeband=(None, None)),
        ]
        siac.accounts = acc_list2
        assert siac.accounts == acc_list2

    @parametrize_with_cases("sync_func", prefix="sync_")
    @parametrize_with_cases("sync_client", prefix="sync_")
    def test_func_errors(
        self,
        account_list,
        unused_tcp_port_factory,
        sync_client,
        sync_func,
    ):
        """Test the function setting."""
        error_type = TypeError if sync_client != sync_func else None
        try:
            if sync_client:
                SIAClient(
                    HOST,
                    unused_tcp_port_factory(),
                    account_list,
                    get_func("sync" if sync_func else "aio"),
                )
            else:
                SIAClientA(
                    HOST,
                    unused_tcp_port_factory(),
                    account_list,
                    get_func("sync" if sync_func else "aio"),
                )
            assert error_type is None
        except Exception as e:
            if isinstance(e, error_type):
                assert True
            else:
                assert False

    @parametrize_with_cases("sync", prefix="sync_")
    async def test_context(self, account_list, unused_tcp_port_factory, sync):
        """Test the context manager functions."""
        if sync:
            with patch.object(SIAClient, "stop") as client:
                with SIAClient(
                    HOST,
                    unused_tcp_port_factory(),
                    account_list,
                    function=get_func("sync"),
                ) as cl:
                    assert cl.accounts == account_list
                client.assert_called_once()

        else:
            with patch.object(SIAClientTCP, "async_stop") as client:
                with patch("asyncio.start_server"):
                    # test Async
                    async with SIAClientA(
                        HOST,
                        port=unused_tcp_port_factory(),
                        accounts=account_list,
                        function=get_func("aio"),
                    ) as cl:
                        assert cl.accounts == account_list
                client.assert_awaited_once()



"""Données XDATA pour pysiaalarm."""

from typing import Dict, Optional

from .data import SIAXDataDict

XDATA: Dict[str, SIAXDataDict] = {
    "A": {
        "characters": "ASCII",
        "description": "A hash of the message that allows the message to be authenticated.",
        "identifier": "A",
        "length": 12,
        "name": "Authentication Hash",
        "value": None,  
    },
    "C": {
        "characters": "ASCII",
        "description": "An identifier for the number of communication paths and link supervision category",
        "identifier": "C",
        "length": 64,
        "name": "Supervision Category",
        "value": None, 
    },
    "H": {
        "characters": "ASCII",
        "description": "Time that event occurred (may be different than message time stamp)",
        "identifier": "H",
        "length": 21,
        "name": "Time of Occurence",
        "value": None,  
    },
    "I": {
        "characters": "Win1252",
        "description": "Alarm text which may be a description of the event or a comment regarding the event.",
        "identifier": "I",
        "length": 256,
        "name": "Alarm Text",
        "value": None,  
    },
    "J": {
        "characters": "ASCII",
        "description": "Manufacturer specific identifier for the path that was used for the communication",
        "identifier": "J",
        "length": 1,
        "name": "Network Path",
        "value": None,  
    },
    "K": {
        "characters": "ASCII",
        "description": "Key exchange request from CSR to PE (up to 256 bits)",
        "identifier": "K",
        "length": 64,
        "name": "Encryption Key",
        "value": None,  
    },
    "L": {
        "characters": "Win1252",
        "description": "Location of event on site",
        "identifier": "L",
        "length": 256,
        "name": "Location",
        "value": None,  
    },
    "M": {
        "characters": "ASCII",
        "description": "MAC address of the PE.",
        "identifier": "M",
        "length": 12,
        "name": "MAC Address",
        "value": None,  
    },
    "N": {
        "characters": "ASCII",
        "description": "Hardware network address associated with the communication on path used.",
        "identifier": "N",
        "length": 128,
        "name": "Network Address",
        "value": None,  
    },
    "O": {
        "characters": "Win1252",
        "description": "Building name.",
        "identifier": "O",
        "length": 256,
        "name": "Building Name",
        "value": None,  
    },
    "P": {
        "characters": "Win1252",
        "description": "contains a message used to support programming or other interactive operations with the receiver",
        "identifier": "P",
        "length": 256,
        "name": "Authentication Hash",
        "value": None,  
    },
    "R": {
        "characters": "Win1252",
        "description": "Room of the event.",
        "identifier": "R",
        "length": 256,
        "name": "Room",
        "value": None,  
    },
    "S": {
        "characters": "Win1252",
        "description": "Site name describing the premises.",
        "identifier": "S",
        "length": 256,
        "name": "Site name",
        "value": None,  
    },
    "T": {
        "characters": "ASCII",
        "description": "Trigger for the event.",
        "identifier": "T",
        "length": 1,
        "name": "Alarm Trigger",
        "value": None,  
    },
    "V": {
        "characters": "Win1252",
        "description": "information about audio or video information that may be associated with the event report.",
        "identifier": "V",
        "length": 256,
        "name": "Verification",
        "value": None,  
    },
    "X": {
        "characters": "ASCII",
        "description": "Location of event, longitude.",
        "identifier": "X",
        "length": 12,
        "name": "Longitude",
        "value": None,  
    },
    "Y": {
        "characters": "ASCII",
        "description": "Location of event, latitude.",
        "identifier": "Y",
        "length": 12,
        "name": "Latitude",
        "value": None,  
    },
    "Z": {
        "characters": "ASCII",
        "description": "Location of event, altitude.",
        "identifier": "Z",
        "length": 12,
        "name": "Altitude",
        "value": None,  
    },
}


#!/usr/bin/python
"""Run a test client."""
import json
import logging
import socket
import sys
import time

from pysiaalarm import CommunicationsProtocol
from .test_utils import create_line_from_test_case, create_random_line

logging.basicConfig(level=logging.DEBUG)
_LOGGER = logging.getLogger(__name__)


def send_messages(config, test_case=None, time_between=5):
    """Create the socket client and start sending messages every 5 seconds, until stopped, or the server disappears."""
    _LOGGER.info("Test client config: %s", config)
    host = config["host"]
    port = config["port"]
    protocol = (
        socket.SOCK_STREAM
        if config["protocol"] == CommunicationsProtocol.TCP
        else socket.SOCK_DGRAM
    )

    with socket.socket(socket.AF_INET, protocol) as sock:
        try:
            sock.connect((host, port))  # connect to the server
        except ConnectionRefusedError:
            _LOGGER.error("Connection refused in test_alarm.py.")
            return

        if test_case:
            message = create_line_from_test_case(config, test_case)
            _LOGGER.debug(f"Sending to server: {message}")
            sock.sendall(message.encode())  # send message
            data = sock.recv(1024).decode()  # receive response
            return
        while True:
            message = create_random_line(config)
            _LOGGER.debug(f"Sending to server: {message}")
            sock.sendall(message.encode())  # send message
            data = sock.recv(1024).decode()  # receive response
            _LOGGER.debug(f"Received from server: {data}")  # show in terminal
            time.sleep(time_between)


if __name__ == "__main__":
    """Run main with a config."""
    _LOGGER.info(sys.argv)
    try:  # sys.argv.index(1)
        file = sys.argv[1]
    except:
        file = "tests//local_config.json"
    with open(file, "r") as f:
        config = json.load(f)
    send_messages(config)



"""Init for aio."""
from .. import (
    CommunicationsProtocol,
    InvalidAccountFormatError,
    InvalidAccountLengthError,
    InvalidKeyFormatError,
    InvalidKeyLengthError,
    SIAAccount,
    SIAEvent,
    __author__,
    __copyright__,
    __license__,
    __version__,
)
from .client import SIAClient



import json
import logging
import time

from pysiaalarm import SIAAccount, SIAClient, SIAEvent

logging.basicConfig(level=logging.DEBUG)

events = []


def func(event: SIAEvent):
    events.append(event)


with open("local_config.json", "r") as f:
    config = json.load(f)
print("Config: ", config)
account = [SIAAccount(config["account_id"], config["key"])]
sleep_time = 1200
with SIAClient(config["host"], config["port"], account, function=func) as client:
    time.sleep(sleep_time)
    counts = client.counts

# for ev in events:
#     print(ev)

print("--------------------------------------------------")
print("Number of events: ", len(events))
print("Counts: ", counts)
print("--------------------------------------------------")



"""Utils for testing pysiaalarm."""
import logging
from binascii import hexlify
from datetime import datetime, timedelta

import random
from Crypto import Random
from Crypto.Cipher import AES

from pysiaalarm.utils import (
    _load_sia_codes,
)
from pysiaalarm.const import IV

logging.basicConfig(level=logging.INFO)
_LOGGER = logging.getLogger(__name__)

BASIC_CONTENT = f"|Nri<zone>/<code>000]<timestamp>"
BASE_LINE = f'<type>"<seq>L0#<account>[<content>'
UNKNOWN_CODE = "ZX"

KEY = "AAAAAAAAAAAAAAAA"
ACCOUNT = "1111"
HOST = "127.0.0.1"
PORT = 7777

ADM_REVERSE_MAP = {"RP": 1602, "WA": 1113}


def _get_crypter(key: bytes):
    """Give back a encrypter/decrypter."""
    return AES.new(key, AES.MODE_CBC, IV)


def crc_calc(msg: str) -> str:
    """Calculate the CRC of the msg."""
    crc = 0
    for letter in str.encode(msg):
        temp = letter
        for _ in range(0, 8):
            temp ^= crc & 1
            crc >>= 1
            if (temp & 1) != 0:
                crc ^= 0xA001
            temp >>= 1
    return ("%x" % crc).upper().zfill(4)


def _create_padded_message(message):
    """Pad the message."""
    fill_size = len(message) + 16 - len(message) % 16
    return message.zfill(fill_size)


def create_random_line(config):
    """Create a full test line with some random values."""
    return create_test_line(
        account="FFFFFFFFF" if random.random() < 0.1 else config["account_id"],
        key=config["key"],
        code=UNKNOWN_CODE if random.random() < 0.1 else _get_random_code(),
        seq=str(random.randint(1000, 9999)),
        time_offset=timedelta(seconds=random.randint(0, 100)),
        alter_crc=random.random() < 0.1,
    )


def create_line_from_test_case(config, tc):
    """Create a full test line."""
    return create_test_line(
        account="FFFFFFFFF" if tc.get("account", False) else config["account_id"],
        key=config["key"],
        code=UNKNOWN_CODE if tc.get("code", False) else _get_random_code(),
        seq=str(random.randint(1000, 9999)),
        time_offset=timedelta(seconds=100 if tc.get("timestamp", False) else 0),
        alter_crc=tc.get("crc", False),
        wrong_event=tc.get("format", False),
    )


def create_test_line(
    account,
    key,
    code,
    seq="7654",
    time_offset=timedelta(0),
    msg_type="SIA-DCS",
    alter_crc=False,
    alter_key=False,
    wrong_event=False,
    use_fixed_time=False,
):
    """Create a test line, with encrytion if key is supplied."""
    if wrong_event:
        return "This is not a SIA Event."
    timestamp = _get_timestamp(time_offset)
    if use_fixed_time:
        timestamp = "_16:04:02,07-09-2020"
    if key:
        if alter_key:
            _LOGGER.debug("Old key: %s", key)
            key = key[:-1] + str(int(key[-1], 16) - 1)
            _LOGGER.debug("New key: %s", key)
    line = _construct_string(msg_type, seq, account, code, timestamp, key)
    if alter_crc:
        crc = ("%04x" % random.randrange(16 ** 4)).upper()
    else:
        crc = crc_calc(line)
    leng = str(int(str(len(line)), 16)).zfill(4)
    return rf"{crc}{leng}{line}"


def _construct_string(msg_type, seq, account, code, timestamp, key=None):
    """Construct the string based on the inputs."""
    return f'"{"*" if key else ""}{msg_type}"{seq}L0#{account}[{_construct_content(msg_type, "0" if code == "RP" else "1", code, timestamp, key)}'


def _construct_content(msg_type, zone, code, timestamp, key=None):
    """Construct the content of the message."""
    cont = f"]{timestamp}"
    if msg_type == "SIA-DCS":
        cont = f"|Nri{zone}/{code}000" + cont
    if msg_type == "ADM-CID":
        # z = f"{int(zone):03d}"
        cont = f"|{ADM_REVERSE_MAP.get(code)} 00 {int(zone):03d}" + cont
    if key:
        return _encrypt_content(key, cont)
    return cont


def _encrypt_content(key, content):
    """Create encrypted content."""
    if not isinstance(key, bytes):
        key = key.encode("utf-8")
    encrypter = _get_crypter(key)
    extra = len(content) % 16
    unencrypted = (16 - extra) * "0" + content
    return (
        hexlify(encrypter.encrypt(unencrypted.encode("utf-8")))
        .decode(encoding="utf-8")
        .upper()
    )


def _get_timestamp(timed: timedelta) -> str:
    """Create a timestamp in the right format."""
    return (datetime.utcnow() - timed).strftime("_%H:%M:%S,%m-%d-%Y")


def _get_random_code() -> str:
    """Get a random code from all codes."""
    codes = [code for code in _load_sia_codes()]
    return random.choice(codes)



import logging
import re

logging.basicConfig(level=logging.DEBUG)
from pysiaalarm import SIAAccount, SIAEvent
from pysiaalarm.utils.regexes import (
    MAIN_MATCHER,
    ENCR_ADM_CONTENT_MATCHER,
    ENCR_SIA_CONTENT_MATCHER,
    SIA_CONTENT_MATCHER,
    ADM_CONTENT_MATCHER,
    OH_MATCHER,
)

# # matcher = re.compile(regex, re.X)
# # print(matcher.pattern)
# main_regex = r"""
# (?P<crc>[A-F0-9]{4})
# (?P<length>[A-F0-9]{4})\"
# (?P<encrypted_flag>[*])?
# (?P<message_type>SIA-DCS|NULL)\"
# (?P<sequence>[0-9]{4})
# (?P<receiver>R[A-F0-9]{1,6})?
# (?P<prefix>L[A-F0-9]{1,6})
# [#]?(?P<account>[A-F0-9]{3,16})?
# [\[]
# (?P<rest>.*)
# """
# MAIN_MATCHER = re.compile(main_regex, re.X)

# # “ti”hh:mm/ time (e.g. ti10:23/).
# # “id”nnn/ user number, if applicable; otherwise not sent (e.g.
# # “ri”nn/ partition no. (e.g. ri12/ or ri3).
# content_regex = r"""
# [#]?(?P<account>[A-F0-9]{3,16})?
# [|]?
# [N]?
# (?:ti)?(?:(?<=ti)(?P<ti>\d{2}:\d{2}))?\/?
# (?:id)?(?:(?<=id)(?P<id>\d*))?\/?
# (?:ri)?(?:(?<=ri)(?P<ri>\d*))?\/?
# (?P<code>[a-zA-z]{2})?
# (?P<message>\w*)
# [\]]
# (?:\[(?:(?<=\[)(?P<extendeddata>\w*)(?=\]))\])?
# [_]?
# (?P<timestamp>[0-9:,-]*)?
# """
# CONTENT_MATCHER = re.compile(content_regex, re.X)

# encr_content_regex = r"""(?:[^\|\[\]]*)[|]?"""
# ENCR_CONTENT_MATCHER = re.compile(encr_content_regex + content_regex, re.X)

# heartbeat_regex = r"""
# ^S
# (?:R)(?:(?<=R)(?P<receiver>\d{4}))
# (?:L)(?:(?<=L)(?P<prefix>\d{4}))
# \s+\w{8}\s+
# \[(?P<id>\w+)\]$
# """

# HEARTBEAT_MATCHER = re.compile(heartbeat_regex, re.X)

sia_lines = [
    r'2E680078"SIA-DCS"6002L0#AAA[|Nri1/CL501]_14:12:04,09-25-2019',
    r'76D80055"*NULL"0000R0L0#AAAB[B4BC8B40D0E6D959D6BEA78E88CC0B2155741A3C44FBB96D476A3E557CAD64D9',
    r'01010051"SIA-DCS"4738R0001L0001[#006969|Nri04/OP001*’NM]DDA2FBB313169D87|#006969',
    r'C4160279"SIA-DCS"5268L0#AAA[.Rr\x1d PaG\'5"�\n\x03��|Nri1/WA000]_08:40:47,07-08-2020',
    r'68370130"*NULL"6327L0#AAA[9F719719F36B05547CD730D4615FE0A4B5BB7A27A9F500741A07C3F7328FC7A1',
    r'43580023"SIA-DCS"0084L0#AAA[#AAA|Nri1/XC12]'
    r'85DF0078"*SIA-DCS"4480L0#EA1984[83F153789366885D5F83DD5A8D19F691DE6602D5D71342E244C040C5D10D89040444068312750F38DF7E63AD3DE8AD5A',
    r'1908002B"SIA-DCS"0000L0#3080[#3080|Nti19:44/id1/ri4/RP]',
    r'43580023"*SIA-DCS"0084L0#AAA[P$WwA!#|#EA1984|Nri1/NL4]_08:38:13,01-07-2021',
    r'2E680078"SIA-DCS"6002L0#AAA[|Nri1/CL501][A38475345]_14:12:04,09-25-2019',
    r"SR0001L0001    006969XX    [ID00000000]",
]

adm_lines = [
    r'87CD0037"ADM-CID"9876R579BDFL789ABC#12345A[#12345A|1110 00 129]_14:12:04,09-25-201'
]

encr_content = [
    r"v V|#EA1984|Nri0/RP0000]_16:57:20,08-11-2020",
    r" HM|#EA1984|Nri0/RP0000]_19:52:37,08-11-2020",
    r"T,o |#EA1984|Nri0/RP0000]_20:44:12,08-11-2020",
    r"_ |#EA1984|Nri0/RP0000]_22:27:02,08-11-2020",
    r"&* |#EA1984|Nri0/RP0000]_05:19:49,08-12-2020",
    r"P$WwA!#|#EA1984|Nri1/NL4]_08:38:13,01-07-2021",
]
unen_content = [
    r"#006969|Nri04/OP001NM]",
]

# for content in unen_content:
#     content_match = SIA_CONTENT_MATCHER.match(content)
#     print("unen_content_match groups", content_match.groupdict())

# for content in encr_content:
#     en_content_match = ENCR_SIA_CONTENT_MATCHER.match(content)
#     print("en_content_match groups", en_content_match.groupdict())

# for line in sia_lines:
#     print("Line ", line)
#     # print(SIAEvent(line))
#     prefix = MAIN_MATCHER.match(line)
#     # re.match(prefix_regex, line, re.X)
#     if prefix:
#         print("Groups", prefix.groupdict())
#         if prefix.group("encrypted_flag"):
#             # if re.search(r"(\*SIA-DCS|\*NULL)", line):
#             encrypted_content = prefix.group("rest")
#             print(encrypted_content)
#             content = ENCR_SIA_CONTENT_MATCHER.match(encrypted_content)
#             if content:
#                 print("Groups ", content.groupdict())
#             # print(
#             #     "Groups ", re.match(regex_encrypted, prefix.group("rest"), re.X).groupdict()
#             # )
#         else:
#             content = SIA_CONTENT_MATCHER.match(prefix.group("rest"))
#             if content:
#                 print("Groups ", content.groupdict())
#     else:
#         heartbeat = OH_MATCHER.match(line)
#         if heartbeat:
#             print("Groups ", heartbeat.groupdict())

for line in adm_lines:
    print("Line ", line)
    # print(SIAEvent(line))
    prefix = MAIN_MATCHER.match(line)
    # re.match(prefix_regex, line, re.X)
    if prefix:
        print("Groups", prefix.groupdict())
        if prefix.group("encrypted_flag"):
            # if re.search(r"(\*SIA-DCS|\*NULL)", line):
            encrypted_content = prefix.group("rest")
            print(encrypted_content)
            content = ENCR_ADM_CONTENT_MATCHER.match(encrypted_content)
            if content:
                print("Groups ", content.groupdict())
            # print(
            #     "Groups ", re.match(regex_encrypted, prefix.group("rest"), re.X).groupdict()
            # )
        else:
            content = ADM_CONTENT_MATCHER.match(prefix.group("rest"))
            if content:
                print("Groups ", content.groupdict())
    else:
        heartbeat = OH_MATCHER.match(line)
        if heartbeat:
            print("Groups ", heartbeat.groupdict())



"""This is the base class with the handling logic for both sia_servers."""
from __future__ import annotations

import logging
from abc import ABC
from collections.abc import Awaitable, Callable

from .account import SIAAccount
from .const import (
    COUNTER_ACCOUNT,
    COUNTER_CODE,
    COUNTER_CRC,
    COUNTER_EVENTS,
    COUNTER_FORMAT,
    COUNTER_TIMESTAMP,
    COUNTER_USER_CODE,
)
from .errors import EventFormatError, NoAccountError
from .event import NAKEvent, OHEvent, SIAEvent, EventsType
from .utils import Counter, ResponseType

_LOGGER = logging.getLogger(__name__)


class BaseSIAServer(ABC):
    """Base class for SIA Server."""

    def __init__(
        self,
        accounts: dict[str, SIAAccount],
        counts: Counter,
        func: Callable[[SIAEvent], None] | None = None,
        async_func: Callable[[SIAEvent], Awaitable[None]] | None = None,
    ):
        """Create a SIA Server.

        Arguments:
            accounts Dict[str, SIAAccount] -- accounts as dict with account_id as key, SIAAccount object as value.  # pylint: disable=line-too-long
            func Callable[[SIAEvent], None] -- Function called for each valid SIA event, that can be matched to a account.  # pylint: disable=line-too-long
            counts Counter -- counter kept by client to give insights in how many errorous EventsType were discarded of each type.  # pylint: disable=line-too-long
        """
        self.accounts = accounts
        self.func = func
        self.async_func = async_func
        self.counts = counts
        self.shutdown_flag = False

    def parse_and_check_event(self, data: bytes) -> EventsType | None:
        """Parse and check the line and create the event, check the account and define the response.

        Args:
            line (str): Line to parse

        Returns:
            SIAEvent: The SIAEvent type of the parsed line.
            ResponseType: The response to send to the alarm.

        """
        line = str.strip(data.decode("ascii", errors="ignore"))
        if not line:
            return None
        self.log_and_count(COUNTER_EVENTS, line=line)
        try:
            event = SIAEvent.from_line(line, self.accounts)
        except NoAccountError as exc:
            self.log_and_count(COUNTER_ACCOUNT, line, exception=exc)
            return NAKEvent()
        except EventFormatError as exc:
            self.log_and_count(COUNTER_FORMAT, line, exception=exc)
            return NAKEvent()

        if isinstance(event, OHEvent):
            return event  # pragma: no cover
        if not event.valid_message:
            self.log_and_count(COUNTER_CRC, event=event)
        elif not event.sia_account:
            self.log_and_count(COUNTER_ACCOUNT, event=event)
        elif event.code_not_found:
            self.log_and_count(COUNTER_CODE, event=event, unknown_code=event.code)  # Passer le code
        elif not event.valid_timestamp:
            self.log_and_count(COUNTER_TIMESTAMP, event=event)
        return event

    async def async_func_wrap(self, event: EventsType | None) -> None:
        """Wrap the user function in a try."""
        if (
            event is None
            or not (isinstance(event, SIAEvent))
            or event.response != ResponseType.ACK
        ):
            return
        self.counts.increment_valid_events()
        try:
            assert self.async_func is not None
            await self.async_func(event)  # type: ignore
        except Exception as exp:  # pylint: disable=broad-except
            self.log_and_count(COUNTER_USER_CODE, event=event, exception=exp)

    def func_wrap(self, event: EventsType | None) -> None:
        """Wrap the user function in a try."""
        if (
            event is None
            or not (isinstance(event, SIAEvent))
            or event.response != ResponseType.ACK
        ):
            return
        self.counts.increment_valid_events()
        try:
            assert self.func is not None
            self.func(event)
        except Exception as exp:  # pylint: disable=broad-except
            self.log_and_count(COUNTER_USER_CODE, event=event, exception=exp)

    def log_and_count(
        self,
        counter: str,
        line: str | None = None,
        event: SIAEvent | None = None,
        exception: Exception | None = None,
        unknown_code: str | None = None,  # Ajouter un paramètre pour le code inconnu
    ) -> None:
        """Log the appropriate line and increment the right counter."""
        if counter == COUNTER_ACCOUNT and exception is not None:
            _LOGGER.warning(
                "There is no account for a encrypted line, line was: %s",
                line,
            )
        if counter == COUNTER_ACCOUNT and event:
            _LOGGER.warning(
                "Unknown or non-existing account (%s) was used by the event: %s",
                event.account,
                event,
            )
        if counter == COUNTER_FORMAT and exception:
            _LOGGER.warning(
                "Last line could not be parsed succesfully. Error message: %s. Line: %s",
                exception.args[0],
                line,
            )
        if counter == COUNTER_USER_CODE and event and exception:
            _LOGGER.warning(
                "Last event: %s, gave error in user function: %s.", event, exception
            )
        if counter == COUNTER_CRC and event:
            _LOGGER.warning(
                "CRC mismatch, ignoring message. Sent CRC: %s, Calculated CRC: %s. Line was %s",
                event.msg_crc,
                event.calc_crc,
                event.full_message,
            )
        if counter == COUNTER_CODE and event:
            _LOGGER.warning(
                "Code not found (%s), replying with DUH to account: %s", unknown_code, event.account
            )
            # Vous pouvez maintenant accéder à unknown_code ici
            _LOGGER.debug(f"Code inconnu détecté pour l'événement {event}: {unknown_code}")
        if counter == COUNTER_TIMESTAMP and event:
            _LOGGER.warning("Event timestamp is no longer valid: %s", event.timestamp)
        if counter == COUNTER_EVENTS and line:
            _LOGGER.debug("Incoming line: %s", line)
        self.counts.increment(counter)


SIA_CODES = {
    "AA": {
        "code": "AA",
        "concerns": "Condition number",
        "description": "An attempt to substitute an alternate alarm panel "
        "for a secure panel has been made",
        "type": "Alarm – Panel Substitution",
    },
    "AB": {
        "code": "AB",
        "concerns": "Zone or point",
        "description": "An event message was not sent due to User action",
        "type": "Abort",
    },
    "AF": {
        "code": "AF",
        "concerns": "Zone or point",
        "description": "Scenario",
        "type": "Arming Event",
    },
    "AN": {
        "code": "AN",
        "concerns": "Zone or point",
        "description": "An analog fire sensor has been restored to normal operation",
        "type": "Analog Restoral",
    },
    "AR": {
        "code": "AR",
        "concerns": "Unused",
        "description": "AC power has been restored",
        "type": "AC Restoral",
    },
    "AS": {
        "code": "AS",
        "concerns": "Zone or point",
        "description": "An analog fire sensor needs to be cleaned or calibrated",
        "type": "Analog Service",
    },
    "AT": {
        "code": "AT",
        "concerns": "Unused",
        "description": "AC power has been failed",
        "type": "AC Trouble",
    },
    "BA": {
        "code": "BA",
        "concerns": "Zone or point",
        "description": "Burglary zone has been violated while armed",
        "type": "Burglary Alarm",
    },
    "BB": {
        "code": "BB",
        "concerns": "Zone or point",
        "description": "Burglary zone has been bypassed",
        "type": "Burglary Bypass",
    },
    "BC": {
        "code": "BC",
        "concerns": "User number",
        "description": "Alarm has been cancelled by authorized user",
        "type": "Burglary Cancel",
    },
    "BD": {
        "code": "BD",
        "concerns": "Zone or point",
        "description": "A non-fire zone has been violated after a Swinger "
        "Shutdown on the zone",
        "type": "Swinger Trouble",
    },
    "BE": {
        "code": "BE",
        "concerns": "Zone or point",
        "description": "A non-fire zone restores to normal from a Swinger "
        "Trouble state",
        "type": "Swinger Trouble Restore",
    },
    "BG": {
        "code": "BG",
        "concerns": "Zone or point",
        "description": "A point assigned to a Cross Point group has gone "
        "into alarm but the Cross Point remained normal",
        "type": "Unverified Event - Burglary",
    },
    "BH": {
        "code": "BH",
        "concerns": "Zone or point",
        "description": "Alarm condition eliminated",
        "type": "Burglary Alarm Restore",
    },
    "BJ": {
        "code": "BJ",
        "concerns": "Zone or point",
        "description": "Trouble condition eliminated",
        "type": "Burglary Trouble Restore",
    },
    "BM": {
        "code": "BM",
        "concerns": "Zone or point",
        "description": "Burglary alarm w/cross point also in alarm - alarm verified",
        "type": "Burglary Alarm - Cross Point",
    },
    "BR": {
        "code": "BR",
        "concerns": "Zone or point",
        "description": "Alarm/trouble condition has been eliminated",
        "type": "Burglary Restoral",
    },
    "BS": {
        "code": "BS",
        "concerns": "Zone or point",
        "description": "Unsafe intrusion detection system condition",
        "type": "Burglary Supervisory",
    },
    "BT": {
        "code": "BT",
        "concerns": "Zone or point",
        "description": "Burglary zone disabled by fault",
        "type": "Burglary Trouble",
    },
    "BU": {
        "code": "BU",
        "concerns": "Zone or point",
        "description": "Zone bypass has been removed",
        "type": "Burglary Unbypass",
    },
    "BV": {
        "code": "BV",
        "concerns": "Area number",
        "description": "A burglary alarm has occurred and been verified "
        "within programmed conditions. (zone or point not "
        "sent)",
        "type": "Burglary Verified",
    },
    "BX": {
        "code": "BX",
        "concerns": "Zone or point",
        "description": "Burglary zone activated during testing",
        "type": "Burglary Test",
    },
    "BZ": {
        "code": "BZ",
        "concerns": "Zone or point",
        "description": "A non-fire Supervisory point has gone missing",
        "type": "Missing Supervision",
    },
    "CA": {
        "code": "CA",
        "concerns": "Area number",
        "description": "System armed automatically",
        "type": "Automatic Closing",
    },
    "CB": {
        "code": "CB",
        "concerns": "Area number",
        "description": "Group has been disarmed automatically",
        "type": "Arming Event",
    },
    "CC": {
        "code": "CC",
        "concerns": "Area number",
        "description": "Unsuccessful arming attempt",
        "type": "Arming Event",
    },
    "CD": {
        "code": "CD",
        "concerns": "Area number",
        "description": "The system has not been armed for a programmed amount of time",
        "type": "Closing Delinquent",
    },
    "CE": {
        "code": "CE",
        "concerns": "User number",
        "description": "Extend closing time",
        "type": "Closing Extend",
    },
    "CF": {
        "code": "CF",
        "concerns": "User number",
        "description": "System armed, some zones not ready",
        "type": "Forced Closing",
    },
    "CG": {
        "code": "CG",
        "concerns": "Area number",
        "description": "System has been partially armed",
        "type": "Close Area",
    },
    "CI": {
        "code": "CI",
        "concerns": "Area number",
        "description": "An area has not been armed at the end of the closing window",
        "type": "Fail to Close",
    },
    "CJ": {
        "code": "CJ",
        "concerns": "User number",
        "description": "An area was armed after the closing window",
        "type": "Late Close",
    },
    "CK": {
        "code": "CK",
        "concerns": "User number",
        "description": "An area was armed before the closing window",
        "type": "Early Close",
    },
    "CL": {
        "code": "CL",
        "concerns": "User number",
        "description": "System armed, normal",
        "type": "Closing Report",
    },
    "CM": {
        "code": "CM",
        "concerns": "Zone or point",
        "description": "A point has gone missing within 2 minutes of closing",
        "type": "Missing Alarm - Recent Closing",
    },
    "CO": {
        "code": "CO",
        "concerns": "Condition number",
        "description": "A command has been sent to an expansion/peripheral device",
        "type": "Command Sent",
    },
    "CP": {
        "code": "CP",
        "concerns": "User number",
        "description": "System armed automatically",
        "type": "Automatic Closing",
    },
    "CQ": {
        "code": "CQ",
        "concerns": "User number",
        "description": "The system was armed from a remote location",
        "type": "Remote Closing",
    },
    "CR": {
        "code": "CR",
        "concerns": "User number",
        "description": "An alarm occurred within five minutes after the "
        "system was closed",
        "type": "Recent Closing",
    },
    "CS": {
        "code": "CS",
        "concerns": "Zone or point",
        "description": "Account has been armed by keyswitch",
        "type": "Closing Keyswitch",
    },
    "CT": {
        "code": "CT",
        "concerns": "Area number",
        "description": "System was not disarmed on time",
        "type": "Late to Open",
    },
    "CW": {
        "code": "CW",
        "concerns": "Area number",
        "description": "Header for a force armed session, forced point msgs may follow",
        "type": "Was Force Armed",
    },
    "CX": {
        "code": "CX",
        "concerns": "Custom Function number",
        "description": "The panel has executed a preprogrammed set of instructions",
        "type": "Custom Function Executed",
    },
    "CZ": {
        "code": "CZ",
        "concerns": "Zone or point",
        "description": "A point, as opposed to a whole area or account, has closed",
        "type": "Point Closing",
    },
    "DA": {
        "code": "DA",
        "concerns": "User number",
        "description": "An access ID has been added to the controller",
        "type": "Card Assigned",
    },
    "DB": {
        "code": "DB",
        "concerns": "User number",
        "description": "An access ID has been deleted from the controller",
        "type": "Card Deleted",
    },
    "DC": {
        "code": "DC",
        "concerns": "Door number",
        "description": "Access to all users prohibited",
        "type": "Access Closed",
    },
    "DD": {
        "code": "DD",
        "concerns": "Door number",
        "description": "Access denied, unknown code",
        "type": "Access Denied",
    },
    "DE": {
        "code": "DE",
        "concerns": "Door number",
        "description": "An access point was opened via a Request to Enter device",
        "type": "Request to Enter",
    },
    "DF": {
        "code": "DF",
        "concerns": "Door number",
        "description": "Door opened without access request",
        "type": "Door Forced",
    },
    "DG": {
        "code": "DG",
        "concerns": "Door number",
        "description": "Door access granted",
        "type": "Access Granted",
    },
    "DH": {
        "code": "DH",
        "concerns": "Door number",
        "description": "An access point in a Door Left Open state has restored",
        "type": "Door Left Open - Restoral",
    },
    "DI": {
        "code": "DI",
        "concerns": "Door number",
        "description": "Access denied because credential has not exited area "
        "before attempting to re-enter same area",
        "type": "Access Denied – Passback",
    },
    "DJ": {
        "code": "DJ",
        "concerns": "Door number",
        "description": "An access point has been forced open in an unarmed area",
        "type": "Door Forced - Trouble",
    },
    "DK": {
        "code": "DK",
        "concerns": "Door number",
        "description": "Access denied, known code",
        "type": "Access Lockout",
    },
    "DL": {
        "code": "DL",
        "concerns": "Door number",
        "description": "An open access point when open time expired in an armed area",
        "type": "Door Left Open - Alarm",
    },
    "DM": {
        "code": "DM",
        "concerns": "Door number",
        "description": "An open access point when open time expired in an unarmed area",
        "type": "Door Left Open - Trouble",
    },
    "DN": {
        "code": "DN",
        "concerns": "Door number",
        "description": "An access point was open when the door cycle time expired",
        "type": "Door Left Open (non-alarm, non-trouble)",
    },
    "DO": {
        "code": "DO",
        "concerns": "Door number",
        "description": "Access to authorized users allowed",
        "type": "Access Open",
    },
    "DP": {
        "code": "DP",
        "concerns": "Door number",
        "description": "An access request was denied because the request is "
        "occurring outside the user’s authorized time "
        "window(s)",
        "type": "Access Denied - Unauthorized Time",
    },
    "DQ": {
        "code": "DQ",
        "concerns": "Door number",
        "description": "An access request was denied because the user was "
        "not authorized in this area when the area was armed",
        "type": "Access Denied - Unauthorized Arming State",
    },
    "DR": {
        "code": "DR",
        "concerns": "Door number",
        "description": "Access alarm/trouble condition eliminated",
        "type": "Door Restoral",
    },
    "DS": {
        "code": "DS",
        "concerns": "Door number",
        "description": "Identifies door for next report",
        "type": "Door Station",
    },
    "DT": {
        "code": "DT",
        "concerns": "Unused",
        "description": "Access system trouble",
        "type": "Access Trouble",
    },
    "DU": {
        "code": "DU",
        "concerns": "Dealer ID",
        "description": "Dealer ID number",
        "type": "Dealer ID",
    },
    "DV": {
        "code": "DV",
        "concerns": "Door number",
        "description": "An access request was denied because the user is not "
        "authorized in this area",
        "type": "Access Denied - Unauthorized Entry Level",
    },
    "DW": {
        "code": "DW",
        "concerns": "Door number",
        "description": "An access request was denied because the doors "
        "associated Interlock point is open",
        "type": "Access Denied - Interlock",
    },
    "DX": {
        "code": "DX",
        "concerns": "Door number",
        "description": "An access point was opened via a Request to Exit device",
        "type": "Request to Exit",
    },
    "DY": {
        "code": "DY",
        "concerns": "Door number",
        "description": "The door’s lock has been engaged",
        "type": "Door Locked",
    },
    "DZ": {
        "code": "DZ",
        "concerns": "Door number",
        "description": "An access request was denied because the door has "
        "been placed in an Access Closed state",
        "type": "Access Denied - Door Secured",
    },
    "EA": {
        "code": "EA",
        "concerns": "Zone or point",
        "description": "An exit zone remained violated at the end of the "
        "exit delay period",
        "type": "Exit Alarm",
    },
    "EE": {
        "code": "EE",
        "concerns": "User number",
        "description": "An exit zone remained violated at the end of the "
        "exit delay period",
        "type": "Exit Error",
    },
    "EJ": {
        "code": "EJ",
        "concerns": "Expansion device number",
        "description": "Expansion device tamper restoral",
        "type": "Expansion Tamper Restore",
    },
    "EM": {
        "code": "EM",
        "concerns": "Expansion device number",
        "description": "Expansion device missing",
        "type": "Expansion Device Missing",
    },
    "EN": {
        "code": "EN",
        "concerns": "Expansion device number",
        "description": "Expansion device communications re-established",
        "type": "Expansion Missing Restore",
    },
    "ER": {
        "code": "ER",
        "concerns": "Expander number",
        "description": "Expansion device trouble eliminated",
        "type": "Expansion Restoral",
    },
    "ES": {
        "code": "ES",
        "concerns": "Expansion device number",
        "description": "Expansion device enclosure tamper",
        "type": "Expansion Device Tamper",
    },
    "ET": {
        "code": "ET",
        "concerns": "Expander number",
        "description": "Expansion device trouble",
        "type": "Expansion Trouble",
    },
    "EX": {
        "code": "EX",
        "concerns": "Device number",
        "description": "A specific reportable condition is detected on an "
        "external device",
        "type": "External Device Condition",
    },
    "EZ": {
        "code": "EZ",
        "concerns": "Point number",
        "description": "A point remained missing at the end of the exit delay period",
        "type": "Missing Alarm - Exit Error",
    },
    "FA": {
        "code": "FA",
        "concerns": "Zone or point",
        "description": "Fire condition detected",
        "type": "Fire Alarm",
    },
    "FB": {
        "code": "FB",
        "concerns": "Zone or point",
        "description": "Zone has been bypassed",
        "type": "Fire Bypass",
    },
    "FC": {
        "code": "FC",
        "concerns": "Zone or point",
        "description": "A Fire Alarm has been cancelled by an authorized person",
        "type": "Fire Cancel",
    },
    "FG": {
        "code": "FG",
        "concerns": "Zone or point",
        "description": "A point assigned to a Cross Point group has gone "
        "into alarm but the Cross Point remained normal",
        "type": "Unverified Event – Fire",
    },
    "FH": {
        "code": "FH",
        "concerns": "Zone or point",
        "description": "Alarm condition eliminated",
        "type": "Fire Alarm Restore",
    },
    "FI": {
        "code": "FI",
        "concerns": "Area number",
        "description": "The transmitter area's fire test has begun",
        "type": "Fire Test Begin",
    },
    "FJ": {
        "code": "FJ",
        "concerns": "Zone or point",
        "description": "Trouble condition eliminated",
        "type": "Fire Trouble Restore",
    },
    "FK": {
        "code": "FK",
        "concerns": "Area number",
        "description": "The transmitter area's fire test has ended",
        "type": "Fire Test End",
    },
    "FL": {
        "code": "FL",
        "concerns": "Zone or point",
        "description": "The fire panel’s sounder was silenced by command",
        "type": "Fire Alarm Silenced",
    },
    "FM": {
        "code": "FM",
        "concerns": "Point number",
        "description": "Fire Alarm with Cross Point also in alarm verifying "
        "the Fire Alarm",
        "type": "Fire Alarm - Cross Point",
    },
    "FQ": {
        "code": "FQ",
        "concerns": "Zone or point",
        "description": "A fire supervisory zone that was in trouble "
        "condition has now restored to normal",
        "type": "Fire Supervisory Trouble Restore",
    },
    "FR": {
        "code": "FR",
        "concerns": "Zone or point",
        "description": "Alarm/trouble condition has been eliminated",
        "type": "Fire Restoral",
    },
    "FS": {
        "code": "FS",
        "concerns": "Zone or point",
        "description": "Unsafe fire detection system condition",
        "type": "Fire Supervisory",
    },
    "FT": {
        "code": "FT",
        "concerns": "Zone or point",
        "description": "Zone disabled by fault",
        "type": "Fire Trouble",
    },
    "FU": {
        "code": "FU",
        "concerns": "Zone or point",
        "description": "Bypass has been removed",
        "type": "Fire Unbypass",
    },
    "FV": {
        "code": "FV",
        "concerns": "Zone or point",
        "description": "A fire supervision zone that was in alarm has "
        "restored to normal",
        "type": "Fire Supervision Restore",
    },
    "FW": {
        "code": "FW",
        "concerns": "Zone or point",
        "description": "A fire supervisory zone is now in a trouble condition",
        "type": "Fire Supervisory Trouble",
    },
    "FX": {
        "code": "FX",
        "concerns": "Zone or point",
        "description": "Fire zone activated during test",
        "type": "Fire Test",
    },
    "FY": {
        "code": "FY",
        "concerns": "Zone or point",
        "description": "A fire point is now logically missing",
        "type": "Missing Fire Trouble",
    },
    "FZ": {
        "code": "FZ",
        "concerns": "Zone or point",
        "description": "A Fire Supervisory point has gone missing",
        "type": "Missing Fire Supervision",
    },
    "GA": {
        "code": "GA",
        "concerns": "Zone or point",
        "description": "Gas alarm condition detected",
        "type": "Gas Alarm",
    },
    "GB": {
        "code": "GB",
        "concerns": "Zone or point",
        "description": "Zone has been bypassed",
        "type": "Gas Bypass",
    },
    "GH": {
        "code": "GH",
        "concerns": "Zone or point",
        "description": "Alarm condition eliminated",
        "type": "Gas Alarm Restore",
    },
    "GJ": {
        "code": "GJ",
        "concerns": "Zone or point",
        "description": "Trouble condition eliminated",
        "type": "Gas Trouble Restore",
    },
    "GR": {
        "code": "GR",
        "concerns": "Zone or point",
        "description": "Alarm/trouble condition has been eliminated",
        "type": "Gas Restoral",
    },
    "GS": {
        "code": "GS",
        "concerns": "Zone or point",
        "description": "Unsafe gas detection system condition",
        "type": "Gas Supervisory",
    },
    "GT": {
        "code": "GT",
        "concerns": "Zone or point",
        "description": "Zone disabled by fault",
        "type": "Gas Trouble",
    },
    "GU": {
        "code": "GU",
        "concerns": "Zone or point",
        "description": "Bypass has been removed",
        "type": "Gas Unbypass",
    },
    "GX": {
        "code": "GX",
        "concerns": "Zone or point",
        "description": "Zone activated during test",
        "type": "Gas Test",
    },
    "HA": {
        "code": "HA",
        "concerns": "Zone or point",
        "description": "Silent alarm, user under duress",
        "type": "Holdup Alarm",
    },
    "HB": {
        "code": "HB",
        "concerns": "Zone or point",
        "description": "Zone has been bypassed",
        "type": "Holdup Bypass",
    },
    "HH": {
        "code": "HH",
        "concerns": "Zone or point",
        "description": "Alarm condition eliminated",
        "type": "Holdup Alarm Restore",
    },
    "HJ": {
        "code": "HJ",
        "concerns": "Zone or point",
        "description": "Trouble condition eliminated",
        "type": "Holdup Trouble Restore",
    },
    "HR": {
        "code": "HR",
        "concerns": "Zone or point",
        "description": "Alarm/trouble condition has been eliminated",
        "type": "Holdup Restoral",
    },
    "HS": {
        "code": "HS",
        "concerns": "Zone or point",
        "description": "Unsafe holdup system condition",
        "type": "Holdup Supervisory",
    },
    "HT": {
        "code": "HT",
        "concerns": "Zone or point",
        "description": "Zone disabled by fault",
        "type": "Holdup Trouble",
    },
    "HU": {
        "code": "HU",
        "concerns": "Zone or point",
        "description": "Bypass has been removed",
        "type": "Holdup Unbypass",
    },
    "IA": {
        "code": "IA",
        "concerns": "Point number",
        "description": "A specific, reportable condition is detected on a device",
        "type": "Equipment Failure Condition",
    },
    "IR": {
        "code": "IR",
        "concerns": "Point number",
        "description": "The equipment condition has been restored to normal",
        "type": "Equipment Fail - Restoral",
    },
    "JA": {
        "code": "JA",
        "concerns": "Area number",
        "description": "Too many unsuccessful attempts have been made to "
        "enter a user ID",
        "type": "User code Tamper",
    },
    "JC": {
        "code": "JC",
        "concerns": "Area number",
        "description": "Break password restore",
        "type": "User code tamper canceled",
    },
    "JD": {
        "code": "JD",
        "concerns": "User number",
        "description": "The date was changed in the transmitter/receiver",
        "type": "Date Changed",
    },
    "JH": {
        "code": "JH",
        "concerns": "User number",
        "description": "The transmitter's holiday schedule has been changed",
        "type": "Holiday Changed",
    },
    "JK": {
        "code": "JK",
        "concerns": "User number",
        "description": "A designated user passcode has not been entered "
        "during a scheduled time window",
        "type": "Latchkey Alert",
    },
    "JL": {
        "code": "JL",
        "concerns": "Unused",
        "description": "The transmitter's log memory has reached its threshold level",
        "type": "Log Threshold",
    },
    "JO": {
        "code": "JO",
        "concerns": "Unused",
        "description": "The transmitter's log memory has overflowed",
        "type": "Log Overflow",
    },
    "JP": {
        "code": "JP",
        "concerns": "User number",
        "description": "A designated user passcode has been used to gain "
        "access to the premises.",
        "type": "User On Premises",
    },
    "JR": {
        "code": "JR",
        "concerns": "Area number",
        "description": "An automatic scheduled event was executed",
        "type": "Schedule Executed",
    },
    "JS": {
        "code": "JS",
        "concerns": "User number",
        "description": "An automatic schedule was changed",
        "type": "Schedule Changed",
    },
    "JT": {
        "code": "JT",
        "concerns": "User number",
        "description": "The time was changed in the transmitter/receiver",
        "type": "Time Changed",
    },
    "JV": {
        "code": "JV",
        "concerns": "User number",
        "description": "A user's code has been changed",
        "type": "User code Changed",
    },
    "JX": {
        "code": "JX",
        "concerns": "User number",
        "description": "A user's code has been removed",
        "type": "User code Deleted",
    },
    "JY": {
        "code": "JY",
        "concerns": "User number",
        "description": "A user’s code has been added",
        "type": "User code Added",
    },
    "JZ": {
        "code": "JZ",
        "concerns": "User number",
        "description": "A user’s authority level has been set",
        "type": "User Level Set",
    },
    "KA": {
        "code": "KA",
        "concerns": "Zone or point",
        "description": "High temperature detected on premise",
        "type": "Heat Alarm",
    },
    "KB": {
        "code": "KB",
        "concerns": "Zone or point",
        "description": "Zone has been bypassed",
        "type": "Heat Bypass",
    },
    "KH": {
        "code": "KH",
        "concerns": "Zone or point",
        "description": "Alarm condition eliminated",
        "type": "Heat Alarm Restore",
    },
    "KJ": {
        "code": "KJ",
        "concerns": "Zone or point",
        "description": "Trouble condition eliminated",
        "type": "Heat Trouble Restore",
    },
    "KR": {
        "code": "KR",
        "concerns": "Zone or point",
        "description": "Alarm/trouble condition has been eliminated",
        "type": "Heat Restoral",
    },
    "KS": {
        "code": "KS",
        "concerns": "Zone or point",
        "description": "Unsafe heat detection system condition",
        "type": "Heat Supervisory",
    },
    "KT": {
        "code": "KT",
        "concerns": "Zone or point",
        "description": "Zone disabled by fault",
        "type": "Heat Trouble",
    },
    "KU": {
        "code": "KU",
        "concerns": "Zone or point",
        "description": "Bypass has been removed",
        "type": "Heat Unbypass",
    },
    "LB": {
        "code": "LB",
        "concerns": "Unused",
        "description": "Begin local programming",
        "type": "Local Program",
    },
    "LD": {
        "code": "LD",
        "concerns": "Unused",
        "description": "Access code incorrect",
        "type": "Local Program Denied",
    },
    "LE": {
        "code": "LE",
        "concerns": "Unused",
        "description": "The listen-in session has been terminated",
        "type": "Listen-in Ended",
    },
    "LF": {
        "code": "LF",
        "concerns": "Unused",
        "description": "The listen-in session with the RECEIVER has begun",
        "type": "Listen-in Begin",
    },
    "LR": {
        "code": "LR",
        "concerns": "Line number",
        "description": "Phone line restored to service",
        "type": "Phone Line Restoral",
    },
    "LS": {
        "code": "LS",
        "concerns": "Unused",
        "description": "Local programming successful",
        "type": "Local Program Success",
    },
    "LT": {
        "code": "LT",
        "concerns": "Line number",
        "description": "Phone line trouble report",
        "type": "Phone Line Trouble",
    },
    "LU": {
        "code": "LU",
        "concerns": "Unused",
        "description": "Local programming unsuccessful",
        "type": "Local Program Fail",
    },
    "LX": {
        "code": "LX",
        "concerns": "Unused",
        "description": "A local programming session has been terminated",
        "type": "Local Programming Ended",
    },
    "MA": {
        "code": "MA",
        "concerns": "Zone or point",
        "description": "Emergency assistance request",
        "type": "Medical Alarm",
    },
    "MB": {
        "code": "MB",
        "concerns": "Zone or point",
        "description": "Zone has been bypassed",
        "type": "Medical Bypass",
    },
    "MH": {
        "code": "MH",
        "concerns": "Zone or point",
        "description": "Alarm condition eliminated",
        "type": "Medical Alarm Restore",
    },
    "MI": {
        "code": "MI",
        "concerns": "Message number",
        "description": "A canned message is being sent",
        "type": "Message",
    },
    "MJ": {
        "code": "MJ",
        "concerns": "Zone or point",
        "description": "Trouble condition eliminated",
        "type": "Medical Trouble Restore",
    },
    "MR": {
        "code": "MR",
        "concerns": "Zone or point",
        "description": "Alarm/trouble condition has been eliminated",
        "type": "Medical Restoral",
    },
    "MS": {
        "code": "MS",
        "concerns": "Zone or point",
        "description": "Unsafe system condition exists",
        "type": "Medical Supervisory",
    },
    "MT": {
        "code": "MT",
        "concerns": "Zone or point",
        "description": "Zone disabled by fault",
        "type": "Medical Trouble",
    },
    "MU": {
        "code": "MU",
        "concerns": "Zone or point",
        "description": "Bypass has been removed",
        "type": "Medical Unbypass",
    },
    "NA": {
        "code": "NA",
        "concerns": "Zone number",
        "description": "There has been no zone activity for a programmed "
        "amount of time",
        "type": "No Activity",
    },
    "NB": {
        "code": "NB",
        "concerns": "Area number",
        "description": "Scenario",
        "type": "Arming Event",
    },
    "NC": {
        "code": "NC",
        "concerns": "Network number",
        "description": "A communications network has a specific reportable condition",
        "type": "Network Condition",
    },
    "ND": {
        "code": "ND",
        "concerns": "Area number",
        "description": "Night mode deactivated - duress",
        "type": "Arming Event",
    },
    "NE": {
        "code": "NE",
        "concerns": "Area number",
        "description": "Unsuccessful Night mode activation attempt",
        "type": "Arming Event",
    },
    "NF": {
        "code": "NF",
        "concerns": "Area number",
        "description": "Some zones/points not ready",
        "type": "Forced Perimeter Arm",
    },
    "NL": {
        "code": "NL",
        "concerns": "Area number",
        "description": "An area has been perimeter armed",
        "type": "Perimeter Armed",
    },
    "NM": {
        "code": "NM",
        "concerns": "Area number",
        "description": "A user defined area has been perimeter armed",
        "type": "Perimeter Armed, User Defined",
    },
    "NO": {
        "code": "NO",
        "concerns": "Area number",
        "description": "Night mode deactivated automatically",
        "type": "Arming Event",
    },
    "NP": {
        "code": "NP",
        "concerns": "Zone or point",
        "description": "Night mode deactivated using device",
        "type": "Arming Event",
    },
    "NR": {
        "code": "NR",
        "concerns": "Network number",
        "description": "A communications network has returned to normal operation",
        "type": "Network Restoral",
    },
    "NS": {
        "code": "NS",
        "concerns": "Zone number",
        "description": "A zone has detected activity after an alert",
        "type": "Activity Resumed",
    },
    "NT": {
        "code": "NT",
        "concerns": "Network number",
        "description": "A communications network has failed",
        "type": "Network Failure",
    },
    "OA": {
        "code": "OA",
        "concerns": "Area number",
        "description": "System has disarmed automatically",
        "type": "Automatic Opening",
    },
    "OB": {
        "code": "OB",
        "concerns": "Area number",
        "description": "Group has been disarmed automatically",
        "type": "Arming Event",
    },
    "OC": {
        "code": "OC",
        "concerns": "User number",
        "description": "Untyped zone cancel",
        "type": "Cancel Report",
    },
    "OG": {
        "code": "OG",
        "concerns": "Area number",
        "description": "System has been partially disarmed",
        "type": "Open Area",
    },
    "OH": {
        "code": "OH",
        "concerns": "User number",
        "description": "An area in alarm was disarmed before the opening window",
        "type": "Early to Open from Alarm",
    },
    "OI": {
        "code": "OI",
        "concerns": "Area number",
        "description": "An area has not been armed at the end of the opening window",
        "type": "Fail to Open",
    },
    "OJ": {
        "code": "OJ",
        "concerns": "User number",
        "description": "An area was disarmed after the opening window",
        "type": "Late Open",
    },
    "OK": {
        "code": "OK",
        "concerns": "User number",
        "description": "An area was disarmed before the opening window",
        "type": "Early Open",
    },
    "OL": {
        "code": "OL",
        "concerns": "User number",
        "description": "An area in alarm was disarmed after the opening window",
        "type": "Late to Open from Alarm",
    },
    "OP": {
        "code": "OP",
        "concerns": "User number",
        "description": "Account was disarmed",
        "type": "Opening Report",
    },
    "OQ": {
        "code": "OQ",
        "concerns": "User number",
        "description": "The system was disarmed from a remote location",
        "type": "Remote Opening",
    },
    "OR": {
        "code": "OR",
        "concerns": "User number",
        "description": "Account in alarm was reset/disarmed",
        "type": "Disarm From Alarm",
    },
    "OS": {
        "code": "OS",
        "concerns": "Zone or point",
        "description": "Account has been disarmed by keyswitch",
        "type": "Opening Keyswitch",
    },
    "OT": {
        "code": "OT",
        "concerns": "User number",
        "description": "System was not armed on time",
        "type": "Late To Close",
    },
    "OU": {
        "code": "OU",
        "concerns": "Output number",
        "description": "An output on a peripheral device or NAC is not functioning",
        "type": "Output State – Trouble",
    },
    "OV": {
        "code": "OV",
        "concerns": "Output number",
        "description": "An output on a peripheral device or NAC is back to "
        "normal operation",
        "type": "Output State – Restore",
    },
    "OZ": {
        "code": "OZ",
        "concerns": "Zone or point",
        "description": "A point, rather than a full area or account, disarmed",
        "type": "Point Opening",
    },
    "PA": {
        "code": "PA",
        "concerns": "Zone or point",
        "description": "Emergency assistance request, manually activated",
        "type": "Panic Alarm",
    },
    "PB": {
        "code": "PB",
        "concerns": "Zone or point",
        "description": "Panic zone has been bypassed",
        "type": "Panic Bypass",
    },
    "PC": {
        "code": "PC",
        "concerns": "User",
        "description": "Photo on demand enabled",
        "type": "Photo on demand enabled",
    },
    "PD": {
        "code": "PD",
        "concerns": "User",
        "description": "Photo on demand disabled",
        "type": "Photo on demand disabled",
    },
    "PE": {
        "code": "PE",
        "concerns": "Unused",
        "description": "Photo by alarm scenarios feature enabled for this hub",
        "type": "Photo by scenario enable",
    },
    "PF": {
        "code": "PF",
        "concerns": "Zone or point",
        "description": "MotionCam lost connection with hub over photos "
        "communication channel",
        "type": "Malfunction",
    },
    "PG": {
        "code": "PG",
        "concerns": "Unused",
        "description": "Photo by alarm scenarios feature disabled for this hub",
        "type": "Photo by scenario disable",
    },
    "PH": {
        "code": "PH",
        "concerns": "Zone or point",
        "description": "Alarm condition eliminated",
        "type": "Panic Alarm Restore",
    },
    "PJ": {
        "code": "PJ",
        "concerns": "Zone or point",
        "description": "Trouble condition eliminated",
        "type": "Panic Trouble Restore",
    },
    "PO": {
        "code": "PO",
        "concerns": "Unused",
        "description": "Connection restored, device in room",
        "type": "Malfunction restore",
    },
    "PR": {
        "code": "PR",
        "concerns": "Zone or point",
        "description": "Alarm/trouble condition has been eliminated",
        "type": "Panic Restoral",
    },
    "PS": {
        "code": "PS",
        "concerns": "Zone or point",
        "description": "Unsafe system condition exists",
        "type": "Panic Supervisory",
    },
    "PT": {
        "code": "PT",
        "concerns": "Zone or point",
        "description": "Zone disabled by fault",
        "type": "Panic Trouble",
    },
    "PU": {
        "code": "PU",
        "concerns": "Zone or point",
        "description": "Panic zone bypass has been removed",
        "type": "Panic Unbypass",
    },
    "QA": {
        "code": "QA",
        "concerns": "Zone or point",
        "description": "Emergency assistance request",
        "type": "Emergency Alarm",
    },
    "QB": {
        "code": "QB",
        "concerns": "Zone or point",
        "description": "Zone has been bypassed",
        "type": "Emergency Bypass",
    },
    "QH": {
        "code": "QH",
        "concerns": "Zone or point",
        "description": "Alarm condition has been eliminated",
        "type": "Emergency Alarm Restore",
    },
    "QJ": {
        "code": "QJ",
        "concerns": "Zone or point",
        "description": "Trouble condition has been eliminated",
        "type": "Emergency Trouble Restore",
    },
    "QR": {
        "code": "QR",
        "concerns": "Zone or point",
        "description": "Alarm/trouble condition has been eliminated",
        "type": "Emergency Restoral",
    },
    "QS": {
        "code": "QS",
        "concerns": "Zone or point",
        "description": "Unsafe system condition exists",
        "type": "Emergency Supervisory",
    },
    "QT": {
        "code": "QT",
        "concerns": "Zone or point",
        "description": "Zone disabled by fault",
        "type": "Emergency Trouble",
    },
    "QU": {
        "code": "QU",
        "concerns": "Zone or point",
        "description": "Bypass has been removed",
        "type": "Emergency Unbypass",
    },
    "RA": {
        "code": "RA",
        "concerns": "Unused",
        "description": "Transmitter failed to communicate with the remote programmer",
        "type": "Remote Programmer Call Failed",
    },
    "RB": {
        "code": "RB",
        "concerns": "Unused",
        "description": "Remote programming session initiated",
        "type": "Remote Program Begin",
    },
    "RC": {
        "code": "RC",
        "concerns": "Relay number",
        "description": "A relay has energized",
        "type": "Relay Close",
    },
    "RD": {
        "code": "RD",
        "concerns": "Unused",
        "description": "Access passcode incorrect",
        "type": "Remote Program Denied",
    },
    "RN": {
        "code": "RN",
        "concerns": "Unused",
        "description": "A TRANSMITTER was reset via a remote programmer",
        "type": "Remote Reset",
    },
    "RO": {
        "code": "RO",
        "concerns": "Relay number",
        "description": "A relay has de-energized",
        "type": "Relay Open",
    },
    "RP": {
        "code": "RP",
        "concerns": "Unused",
        "description": "Automatic communication test report",
        "type": "Automatic Test",
    },
    "RR": {
        "code": "RR",
        "concerns": "Unused",
        "description": "System lost power, is now restored",
        "type": "Power Up",
    },
    "RS": {
        "code": "RS",
        "concerns": "Unused",
        "description": "Remote programming successful",
        "type": "Remote Program Success",
    },
    "RT": {
        "code": "RT",
        "concerns": "Line number",
        "description": "Dialer data lost, transmission error",
        "type": "Data Lost",
    },
    "RU": {
        "code": "RU",
        "concerns": "Unused",
        "description": "Remote programming unsuccessful",
        "type": "Remote Program Fail",
    },
    "RX": {
        "code": "RX",
        "concerns": "User number",
        "description": "Manual communication test report",
        "type": "Manual Test",
    },
    "RY": {
        "code": "RY",
        "concerns": "Zone or point",
        "description": "Test signal(s) indicates abnormal condition(s) exist",
        "type": "Test Off Normal",
    },
    "SA": {
        "code": "SA",
        "concerns": "Zone or point",
        "description": "Sprinkler flow condition exists",
        "type": "Sprinkler Alarm",
    },
    "SB": {
        "code": "SB",
        "concerns": "Zone or point",
        "description": "Sprinkler zone has been bypassed",
        "type": "Sprinkler Bypass",
    },
    "SC": {
        "code": "SC",
        "concerns": "Condition number",
        "description": "An expansion/peripheral device is reporting a new "
        "condition or state change",
        "type": "Change of State",
    },
    "SH": {
        "code": "SH",
        "concerns": "Zone or point",
        "description": "Alarm condition eliminated",
        "type": "Sprinkler Alarm Restore",
    },
    "SJ": {
        "code": "SJ",
        "concerns": "Zone or point",
        "description": "Trouble condition eliminated",
        "type": "Sprinkler Trouble Restore",
    },
    "SM": {
        "code": "SM",
        "concerns": "Zone or point",
        "description": "LeakProtect Malfunction",
        "type": "Malfunction",
    },
    "SR": {
        "code": "SR",
        "concerns": "Zone or point",
        "description": "Alarm/trouble condition has been eliminated",
        "type": "Sprinkler Restoral",
    },
    "SS": {
        "code": "SS",
        "concerns": "Zone or point",
        "description": "Unsafe sprinkler system condition",
        "type": "Sprinkler Supervisory",
    },
    "ST": {
        "code": "ST",
        "concerns": "Zone or point",
        "description": "Zone disabled by fault",
        "type": "Sprinkler Trouble",
    },
    "SU": {
        "code": "SU",
        "concerns": "Zone or point",
        "description": "Sprinkler zone bypass has been removed",
        "type": "Sprinkler Unbypass",
    },
    "TA": {
        "code": "TA",
        "concerns": "Zone or point",
        "description": "Alarm equipment enclosure opened",
        "type": "Tamper Alarm",
    },
    "TB": {
        "code": "TB",
        "concerns": "Zone or point",
        "description": "Tamper detection has been bypassed",
        "type": "Tamper Bypass",
    },
    "TC": {
        "code": "TC",
        "concerns": "Unused",
        "description": "All point tested",
        "type": "All Points Tested",
    },
    "TE": {
        "code": "TE",
        "concerns": "Unused",
        "description": "Communicator restored to operation",
        "type": "Test End",
    },
    "TH": {
        "code": "TH",
        "concerns": "Unused",
        "description": "An Expansion Device’s tamper switch restores to "
        "normal from an Alarm state",
        "type": "Tamper Alarm Restore",
    },
    "TJ": {
        "code": "TJ",
        "concerns": "Unused",
        "description": "An Expansion Device’s tamper switch restores to "
        "normal from a Trouble state",
        "type": "Tamper Trouble Restore",
    },
    "TP": {
        "code": "TP",
        "concerns": "Point number",
        "description": "This point was tested during a Walk Test",
        "type": "Walk Test Point",
    },
    "TR": {
        "code": "TR",
        "concerns": "Zone or point",
        "description": "Alarm equipment enclosure has been closed",
        "type": "Tamper Restoral",
    },
    "TS": {
        "code": "TS",
        "concerns": "Unused",
        "description": "Communicator taken out of operation",
        "type": "Test Start",
    },
    "TT": {
        "code": "TT",
        "concerns": "Zone or point",
        "description": "Equipment enclosure opened in disarmed state",
        "type": "Tamper Trouble",
    },
    "TU": {
        "code": "TU",
        "concerns": "Zone or point",
        "description": "Tamper detection bypass has been removed",
        "type": "Tamper Unbypass",
    },
    "TW": {
        "code": "TW",
        "concerns": "Unused",
        "description": "Area watch feature has been activated",
        "type": "Area Watch Start",
    },
    "TX": {
        "code": "TX",
        "concerns": "Unused",
        "description": "An unspecified (manual or automatic) communicator test",
        "type": "Test Report",
    },
    "TZ": {
        "code": "TZ",
        "concerns": "Unused",
        "description": "Area watch feature has been deactivated",
        "type": "Area Watch End",
    },
    "UA": {
        "code": "UA",
        "concerns": "Zone or point",
        "description": "Alarm condition from zone of unknown type",
        "type": "Untyped Zone Alarm",
    },
    "UB": {
        "code": "UB",
        "concerns": "Zone or point",
        "description": "Zone of unknown type has been bypassed",
        "type": "Untyped Zone Bypass",
    },
    "UG": {
        "code": "UG",
        "concerns": "Zone or point",
        "description": "A point assigned to a Cross Point group has gone "
        "into alarm but the Cross Point remained normal",
        "type": "Unverified Event – Untyped",
    },
    "UH": {
        "code": "UH",
        "concerns": "Zone or point",
        "description": "Alarm condition eliminated",
        "type": "Untyped Alarm Restore",
    },
    "UJ": {
        "code": "UJ",
        "concerns": "Zone or point",
        "description": "Trouble condition eliminated",
        "type": "Untyped Trouble Restore",
    },
    "UR": {
        "code": "UR",
        "concerns": "Zone or point",
        "description": "Alarm/trouble condition eliminated from zone of unknown type",
        "type": "Untyped Zone Restoral",
    },
    "US": {
        "code": "US",
        "concerns": "Zone or point",
        "description": "Unsafe condition from zone of unknown type",
        "type": "Untyped Zone Supervisory",
    },
    "UT": {
        "code": "UT",
        "concerns": "Zone or point",
        "description": "Trouble condition from zone of unknown type",
        "type": "Untyped Zone Trouble",
    },
    "UU": {
        "code": "UU",
        "concerns": "Zone or point",
        "description": "Bypass on zone of unknown type has been removed",
        "type": "Untyped Zone Unbypass",
    },
    "UX": {
        "code": "UX",
        "concerns": "Unused",
        "description": "An undefined alarm condition has occurred",
        "type": "Undefined",
    },
    "UY": {
        "code": "UY",
        "concerns": "Zone or point",
        "description": "A point or device which was not armed is now logically missing",
        "type": "Untyped Missing Trouble",
    },
    "UZ": {
        "code": "UZ",
        "concerns": "Zone or point",
        "description": "A point or device which was armed is now logically missing",
        "type": "Untyped Missing Alarm",
    },
    "VI": {
        "code": "VI",
        "concerns": "Printer number",
        "description": "TRANSMITTER or RECEIVER paper in",
        "type": "Printer Paper In",
    },
    "VO": {
        "code": "VO",
        "concerns": "Printer number",
        "description": "TRANSMITTER or RECEIVER paper out",
        "type": "Printer Paper Out",
    },
    "VR": {
        "code": "VR",
        "concerns": "Printer number",
        "description": "TRANSMITTER or RECEIVER trouble restored",
        "type": "Printer Restore",
    },
    "VT": {
        "code": "VT",
        "concerns": "Printer number",
        "description": "TRANSMITTER or RECEIVER trouble",
        "type": "Printer Trouble",
    },
    "VX": {
        "code": "VX",
        "concerns": "Printer number",
        "description": "TRANSMITTER or RECEIVER test",
        "type": "Printer Test",
    },
    "VY": {
        "code": "VY",
        "concerns": "Unused",
        "description": "RECEIVER’S printer is now online",
        "type": "Printer Online",
    },
    "VZ": {
        "code": "VZ",
        "concerns": "Unused",
        "description": "RECEIVER’S printer is now offline",
        "type": "Printer Offline",
    },
    "WA": {
        "code": "WA",
        "concerns": "Zone or point",
        "description": "Water detected at protected premises",
        "type": "Water Alarm",
    },
    "WB": {
        "code": "WB",
        "concerns": "Zone or point",
        "description": "Water detection has been bypassed",
        "type": "Water Bypass",
    },
    "WH": {
        "code": "WH",
        "concerns": "Zone or point",
        "description": "Water alarm condition eliminated",
        "type": "Water Alarm Restore",
    },
    "WJ": {
        "code": "WJ",
        "concerns": "Zone or point",
        "description": "Water trouble condition eliminated",
        "type": "Water Trouble Restore",
    },
    "WR": {
        "code": "WR",
        "concerns": "Zone or point",
        "description": "Water alarm/trouble condition has been eliminated",
        "type": "Water Restoral",
    },
    "WS": {
        "code": "WS",
        "concerns": "Zone or point",
        "description": "Water unsafe water detection system condition",
        "type": "Water Supervisory",
    },
    "WT": {
        "code": "WT",
        "concerns": "Zone or point",
        "description": "Water zone disabled by fault",
        "type": "Water Trouble",
    },
    "WU": {
        "code": "WU",
        "concerns": "Zone or point",
        "description": "Water detection bypass has been removed",
        "type": "Water Unbypass",
    },
    "XA": {
        "code": "XA",
        "concerns": "Unused",
        "description": "CS RECEIVER has received an event from a non-existent account",
        "type": "Extra Account Report",
    },
    "XC": {
        "code": "XC",
        "concerns": "Unused",
        "description": "Connection restored, device in room",
        "type": "Malfunction restore",
    },
    "XE": {
        "code": "XE",
        "concerns": "Point number",
        "description": "Panel has sensed an extra point not specified for this site",
        "type": "Extra Point",
    },
    "XF": {
        "code": "XF",
        "concerns": "Point number",
        "description": "Panel has sensed an extra RF point not specified for this site",
        "type": "Extra RF Point",
    },
    "XH": {
        "code": "XH",
        "concerns": "Receiver number",
        "description": "A radio device is no longer detecting RF Interference",
        "type": "RF Interference Restoral",
    },
    "XI": {
        "code": "XI",
        "concerns": "Zone or point",
        "description": "A user has reset a sensor",
        "type": "Sensor Reset",
    },
    "XJ": {
        "code": "XJ",
        "concerns": "Receiver number",
        "description": "A Tamper condition at a premises RF Receiver has been restored",
        "type": "RF Receiver Tamper Restoral",
    },
    "XL": {
        "code": "XL",
        "concerns": "Receiver number",
        "description": "The RF signal strength of a reported event is below "
        "minimum level",
        "type": "Low Received Signal Strength/Connection Lost",
    },
    "XM": {
        "code": "XM",
        "concerns": "Zone or point",
        "description": "Missing Alarm verified by Cross Point in Alarm (or missing)",
        "type": "Missing Alarm - Cross Point",
    },
    "XQ": {
        "code": "XQ",
        "concerns": "Receiver number",
        "description": "A radio device is detecting RF Interference",
        "type": "RF Interference",
    },
    "XR": {
        "code": "XR",
        "concerns": "Zone or point",
        "description": "Low battery has been corrected",
        "type": "Transmitter Battery Restoral",
    },
    "XS": {
        "code": "XS",
        "concerns": "Receiver number",
        "description": "A Tamper condition at a premises receiver is detected",
        "type": "RF Receiver Tamper",
    },
    "XT": {
        "code": "XT",
        "concerns": "Zone or point",
        "description": "Low battery in wireless transmitter",
        "type": "Transmitter Battery Trouble",
    },
    "XW": {
        "code": "XW",
        "concerns": "Zone or point",
        "description": "A point was forced out of the system at arm time",
        "type": "Forced Point",
    },
    "XX": {
        "code": "XX",
        "concerns": "Unused",
        "description": "A specific test from a panel was not received",
        "type": "Fail to Test",
    },
    "YA": {
        "code": "YA",
        "concerns": "Unused",
        "description": "A trouble condition has been detected on a Local "
        "Bell, Siren, or Annunciator",
        "type": "Bell Fault",
    },
    "YB": {
        "code": "YB",
        "concerns": "Line card number",
        "description": "Percent of time receiver's line card is on-line",
        "type": "Busy Seconds",
    },
    "YC": {
        "code": "YC",
        "concerns": "Unused",
        "description": "RECEIVER and TRANSMITTER",
        "type": "Communications Fail",
    },
    "YD": {
        "code": "YD",
        "concerns": "Line card number",
        "description": "A line card identified by the passed address is in trouble",
        "type": "Receiver Line Card Trouble",
    },
    "YE": {
        "code": "YE",
        "concerns": "Line card number",
        "description": "A line card identified by the passed address is restored",
        "type": "Receiver Line Card Restored",
    },
    "YF": {
        "code": "YF",
        "concerns": "Unused",
        "description": "System data corrupted",
        "type": "Parameter Checksum Fail",
    },
    "YG": {
        "code": "YG",
        "concerns": "Unused",
        "description": "A TRANSMITTER’S parameters have been changed",
        "type": "Parameter Changed",
    },
    "YH": {
        "code": "YH",
        "concerns": "Unused",
        "description": "A trouble condition has been restored on a Local "
        "Bell, Siren, or Annunciator",
        "type": "Bell Restored",
    },
    "YI": {
        "code": "YI",
        "concerns": "Unused",
        "description": "An Expansion Device has detected an overcurrent condition",
        "type": "Overcurrent Trouble",
    },
    "YJ": {
        "code": "YJ",
        "concerns": "Unused",
        "description": "An Expansion Device has restored from an overcurrent condition",
        "type": "Overcurrent Restore",
    },
    "YK": {
        "code": "YK",
        "concerns": "Unused",
        "description": "TRANSMITTER has resumed communication with a RECEIVER",
        "type": "Communications Restoral",
    },
    "YM": {
        "code": "YM",
        "concerns": "Unused",
        "description": "TRANSMITTER/RECEIVER battery is missing",
        "type": "System Battery Missing",
    },
    "YN": {
        "code": "YN",
        "concerns": "Unused",
        "description": "TRANSMITTER has sent a packet with invalid data",
        "type": "Invalid Report",
    },
    "YO": {
        "code": "YO",
        "concerns": "Unused",
        "description": "An unknown message was received from automation or the printer",
        "type": "Unknown Message",
    },
    "YP": {
        "code": "YP",
        "concerns": "Unused",
        "description": "TRANSMITTER/RECEIVER has a problem with the power supply",
        "type": "Power Supply Trouble",
    },
    "YQ": {
        "code": "YQ",
        "concerns": "Unused",
        "description": "TRANSMITTER’S/RECEIVER’S power supply has been restored",
        "type": "Power Supply Restored",
    },
    "YR": {
        "code": "YR",
        "concerns": "Unused",
        "description": "Low battery has been corrected",
        "type": "System Battery Restoral",
    },
    "YS": {
        "code": "YS",
        "concerns": "Unused",
        "description": "RECEIVER and TRANSMITTER",
        "type": "Communications Trouble",
    },
    "YT": {
        "code": "YT",
        "concerns": "Unused",
        "description": "Low battery in control/communicator",
        "type": "System Battery Trouble",
    },
    "YU": {
        "code": "YU",
        "concerns": "Condition number",
        "description": "An expansion/peripheral device is reporting a diagnostic error",
        "type": "Diagnostic Error",
    },
    "YW": {
        "code": "YW",
        "concerns": "Unused",
        "description": "The TRANSMITTER created an internal reset",
        "type": "Watchdog Reset",
    },
    "YX": {
        "code": "YX",
        "concerns": "Unused",
        "description": "A TRANSMITTER/RECEIVER needs service",
        "type": "Service Required",
    },
    "YY": {
        "code": "YY",
        "concerns": "Unused",
        "description": "This is a header for an account status report transmission",
        "type": "Status Report",
    },
    "YZ": {
        "code": "YZ",
        "concerns": "Mfr defined",
        "description": "Required TRANSMITTER / RECEIVER service completed",
        "type": "Service Completed",
    },
    "ZA": {
        "code": "ZA",
        "concerns": "Zone or point",
        "description": "Low temperature detected at premises",
        "type": "Freeze Alarm",
    },
    "ZB": {
        "code": "ZB",
        "concerns": "Zone or point",
        "description": "Low temperature detection has been bypassed",
        "type": "Freeze Bypass",
    },
    "ZH": {
        "code": "ZH",
        "concerns": "Zone or point",
        "description": "Alarm condition eliminated",
        "type": "Freeze Alarm Restore",
    },
    "ZJ": {
        "code": "ZJ",
        "concerns": "Zone or point",
        "description": "Trouble condition eliminated",
        "type": "Freeze Trouble Restore",
    },
    "ZR": {
        "code": "ZR",
        "concerns": "Zone or point",
        "description": "Alarm/trouble condition has been eliminated",
        "type": "Freeze Restoral",
    },
    "ZS": {
        "code": "ZS",
        "concerns": "Zone or point",
        "description": "Unsafe freeze detection system condition",
        "type": "Freeze Supervisory",
    },
    "ZT": {
        "code": "ZT",
        "concerns": "Zone or point",
        "description": "Zone disabled by fault",
        "type": "Freeze Trouble",
    },
    "ZU": {
        "code": "ZU",
        "concerns": "Zone or point",
        "description": "Low temperature detection bypass removed",
        "type": "Freeze Unbypass",
    },
    "ZY": {
        "code": "ZY",
        "concerns": "Hub",
        "description": "Device was switched on",
        "type": "Common Event",
    },
    "ZZ": {
        "code": "ZZ",
        "concerns": "Hub",
        "description": "Device was switched off",
        "type": "Common Event",
    },
    "AE": {
        "code": "AE",
        "concerns": "Hub/Device", # Estimation basée sur la description
        "description": "hub switched to battery saving mode / ended call from company [...] due to error", # Note: Apparaît 2 fois avec des descriptions différentes dans le CSV
        "type": "Malfunction",
    },
    "AG": {
        "code": "AG",
        "concerns": "Call/Device", # Estimation basée sur la description
        "description": "ended call to company [...] via [device] in [room]",
        "type": "Common Event",
    },
    "AH": {
        "code": "AH",
        "concerns": "Call/Device", # Estimation basée sur la description
        "description": "started call to company [...] via [device] in [room]",
        "type": "Common Event",
    },
    "AI": {
        "code": "AI",
        "concerns": "Call/Device", # Estimation basée sur la description
        "description": "ended call to company [...] via [device] in [room] due to error",
        "type": "Malfunction",
    },
    "AY": {
        "code": "AY",
        "concerns": "Hub", # Estimation basée sur la description
        "description": "hub exited battery saving mode",
        "type": "Malfunction Restore",
    },
    "HV": {
        "code": "HV",
        "concerns": "System", # Estimation basée sur la description
        "description": "Hold-up alarm is confirmed",
        "type": "Alarm",
    },
    "KG": {
        "code": "KG",
        "concerns": "Keypad/User", # Estimation basée sur la description
        "description": "keypad unlocked on expiration of auto-lock time / keypad unlocked from the app by [user]", # Note: Apparaît 2 fois avec des descriptions similaires
        "type": "Malfunction Restore",
    },
    "RL": {
        "code": "RL",
        "concerns": "Keypad/Switch", # Estimation basée sur la description
        "description": "keypad locked due to unauthorized access attempt / arm switch locked due to device lid being open", # Note: Apparaît 2 fois avec des descriptions différentes
        "type": "Malfunction",
    },
    "RG": {
        "code": "RG",
        "concerns": "Switch/User", # Estimation basée sur la description
        "description": "arm switch unlocked from the app by [user]",
        "type": "Malfunction Restore",
    },
    
}



"""This is the class for the actual TCP handler override of the handle method."""
from __future__ import annotations

import logging
from abc import abstractmethod
from socketserver import BaseRequestHandler

from ..event import SIAEvent

_LOGGER = logging.getLogger(__name__)


class BaseSIAHandler(BaseRequestHandler):
    """Base case for Request handling."""

    _received_data = "".encode()

    def handle_raw_line(self, raw: bytes) -> None:
        """Handle the line."""
        while len(raw) > 0:
            splitter = raw.find(b"\r")
            if splitter == -1:
                data = raw
                raw = b""
            else:  # pragma: no cover
                data = raw[1:splitter]
                raw = raw[splitter + 1 :]
            # decoded_line = line.decode("ascii", errors="ignore")
            event = self.server.parse_and_check_event(data)  # type: ignore
            self.respond(event)
            self.server.func_wrap(event)  # type: ignore

    @abstractmethod
    def respond(self, event: SIAEvent) -> None:
        """Abstract method for responding."""


class SIATCPHandler(BaseSIAHandler):
    """Class for TCP Handling."""

    def handle(self) -> None:
        """Overwritten method for the RequestHandler."""
        while True and not self.server.shutdown_flag:  # type: ignore # pragma: no cover
            raw = self.request.recv(1024)
            if not raw:
                break
            raw = bytearray(raw)
            self.handle_raw_line(raw)

    def respond(self, event: SIAEvent) -> None:
        """Respond to the event."""
        try:
            self.request.sendall(event.create_response())
        except Exception as exp:  # pragma: no cover # pylint: disable=broad-except
            _LOGGER.error(
                "Exception caught while responding to event: %s, exception: %s",
                event.create_response(),
                exp,
            )


class SIAUDPHandler(BaseSIAHandler):
    """Class for UDP Handling."""

    def handle(self) -> None:
        """Overwritten method for the RequestHandler."""
        if not self.server.shutdown_flag:  # type: ignore # pragma: no cover
            raw = self.request[0]
            # socket = self.request[1]
            if raw:
                raw = bytearray(raw)
                self.handle_raw_line(raw)

    def respond(self, event: SIAEvent) -> None:
        """Respond to the event."""
        try:
            self.request[1].sendto(
                event.create_response(),
                self.client_address,
            )
        except Exception as exp:  # pragma: no cover # pylint: disable=broad-except
            _LOGGER.error(
                "Exception caught while responding to event: %s, exception: %s",
                event.create_response(),
                exp,
            )



"""This is a the main class for the SIA Client."""
from __future__ import annotations

import asyncio
from abc import abstractmethod
import logging
from collections.abc import Awaitable, Callable
from types import TracebackType
from typing import Any, Type

from ..account import SIAAccount
from ..base_client import BaseSIAClient
from ..event import SIAEvent
from ..utils import CommunicationsProtocol
from .server import SIAServerTCP, SIAServerUDP

_LOGGER = logging.getLogger(__name__)


class SIAClient(BaseSIAClient):
    """Class for Async SIA Client."""

    protocol: CommunicationsProtocol

    def __new__(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> SIAClient:
        """Create the SIA Client object."""
        subclass_map = {
            subclass.protocol: subclass for subclass in cls.__subclasses__()
        }
        protocol = kwargs.pop("protocol", CommunicationsProtocol.TCP)
        subclass = subclass_map[protocol]
        instance = super(SIAClient, subclass).__new__(subclass)
        return instance

    def __init__(
        self,
        host: str,
        port: int,
        accounts: list[SIAAccount],
        function: Callable[[SIAEvent], Awaitable[None]],
        **kwargs: Any,
    ):
        """Create the asynchronous SIA Client object.

        Arguments:
            host {str} -- Host to run the server on, usually would be ""
            port {int} -- The port the server listens to.
            accounts {List[SIAAccount]} -- List of SIA Accounts to add.
            function {Callable[[SIAEvent], Awaitable[None]]} -- The async function that gets called for each event.  # pylint: disable=line-too-long
            protocol {CommunicationsProtocol Enum} -- CommunicationsProtocol to use, TCP or UDP.

        """
        if not asyncio.iscoroutinefunction(function):
            raise TypeError("Function should be a coroutine, create with async def.")
        BaseSIAClient.__init__(self, host, port, accounts, self.protocol)
        self._func = function

    async def __aenter__(self, **kwargs: Any) -> SIAClient:
        """Start with as context manager."""
        await self.async_start(**kwargs)
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_val: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        """End as context manager."""
        await self.async_stop()
        return True

    @abstractmethod
    async def async_start(self, **kwargs: Any) -> None:
        """Start the asynchronous SIA server."""
        pass

    @abstractmethod
    async def async_stop(self) -> None:
        """Stop the asynchronous SIA server."""
        pass

    async def start(self, **kwargs: Any) -> None:
        """Use async_start instead, this method is depricated and will be removed."""
        _LOGGER.warning(
            "This method is depricated and will be removed, use async_start instead."
        )
        await self.async_start(**kwargs)

    async def stop(self) -> None:
        """Use async_stop instead, this method is depricated and will be removed."""
        _LOGGER.warning(
            "This method is depricated and will be removed, use async_stop instead."
        )
        await self.async_stop()


class SIAClientTCP(SIAClient):
    """TCP subclass."""

    protocol = CommunicationsProtocol.TCP

    def __init__(
        self,
        host: str,
        port: int,
        accounts: list[SIAAccount],
        function: Callable[[SIAEvent], Awaitable[None]],
        **kwargs: Any,
    ) -> None:
        """Create the TCP SIA Client object."""
        super().__init__(host, port, accounts, function)
        self.task: asyncio.Task | None = None
        self.sia_server: SIAServerTCP = SIAServerTCP(
            self._accounts, self._func, self._counts
        )

    async def async_start(self, **kwargs: Any) -> None:
        """Start the asynchronous SIA TCP server.

        The rest of the arguments are passed directly to asyncio.start_server().
        """
        _LOGGER.debug("Starting SIA.")
        coro = asyncio.start_server(
            self.sia_server.handle_line, self._host, self._port, **kwargs
        )
        self.task = asyncio.create_task(coro)

    async def async_stop(self) -> None:
        """Stop the asynchronous SIA TCP server."""
        _LOGGER.debug("Stopping SIA.")
        self.sia_server.shutdown_flag = True
        await asyncio.gather(self.task)  # type: ignore


class SIAClientUDP(SIAClient):
    """UDP Subclass."""

    protocol = CommunicationsProtocol.UDP

    def __init__(
        self,
        host: str,
        port: int,
        accounts: list[SIAAccount],
        function: Callable[[SIAEvent], Awaitable[None]],
        **kwargs: Any,
    ) -> None:
        """Create the UDP SIA Client object."""
        super().__init__(host, port, accounts, function)
        self.transport: asyncio.BaseTransport | None = None
        self.dgprotocol: asyncio.BaseProtocol | None = None

    async def async_start(self, **kwargs: Any) -> None:
        """Start the asynchronous SIA UDP server.

        The rest of the arguments are passed directly to create_datagram_endpoint().
        """
        _LOGGER.debug("Starting SIA.")
        loop = asyncio.get_running_loop()
        self.transport, self.dgprotocol = await loop.create_datagram_endpoint(
            lambda: SIAServerUDP(self._accounts, self._func, self._counts),
            local_addr=(self._host, self._port),
            **kwargs,
        )

    async def async_stop(self) -> None:
        """Stop the asynchronous SIA UDP server."""
        _LOGGER.debug("Stopping SIA.")
        if self.transport is not None:
            self.transport.close()



"""Errors for SIA Server."""


class InvalidKeyFormatError(Exception):
    """Error for when the key is not a hex string."""


class InvalidKeyLengthError(Exception):
    """Error for when the key is not the right amount of characters."""


class InvalidAccountFormatError(Exception):
    """Error for when the account is not a hex string."""


class InvalidAccountLengthError(Exception):
    """Error for when the key is not the right amount of characters."""


class EventFormatError(Exception):
    """Error for when a event is incorrectly formatted."""


class EventCRCError(Exception):
    """Error for when a event hass a mismatched CRC."""


class NoAccountError(Exception):
    """Error for when a account is not present."""



"""This is a the main class for the SIA Client."""
from __future__ import annotations

from abc import ABC

from .account import SIAAccount
from .base_server import BaseSIAServer
from .utils import CommunicationsProtocol, Counter


class BaseSIAClient(ABC):
    """Base class for SIA Client."""

    def __init__(
        self,
        host: str,
        port: int,
        accounts: list[SIAAccount],
        protocol: CommunicationsProtocol = CommunicationsProtocol.TCP,
    ):
        """Create the SIA Client object.

        Arguments:
            host {str} -- Host to run the server on, usually would be ""
            port {int} -- The port the server listens to.
            accounts {List[SIAAccount]} -- List of SIA Accounts to add.
            # function {Callable[[SIAEvent], None]} -- The function that gets called for each event.
            protocol {CommunicationsProtocol Enum} -- CommunicationsProtocol to use, TCP or UDP.

        """
        self._host = host
        self._port = port
        self.protocol = protocol

        self.sia_server: BaseSIAServer | None = None
        self._accounts: dict[str, SIAAccount]
        self.accounts = accounts

        self._counts = Counter()

    @property
    def accounts(self) -> list[SIAAccount]:
        """Return accounts list, ignoring internal structure.

        Returns:
            List[SIAAccount]: List with SIAAccounts

        """
        return list(self._accounts.values())

    @accounts.setter
    def accounts(self, accounts: list[SIAAccount]) -> None:
        """Set the accounts to monitor.

        Args:
            accounts (List[SIAAccount]): List of SIAAccounts to monitor.

        """
        self._accounts = {a.account_id: a for a in accounts}
        if self.sia_server:
            self.sia_server.accounts = self._accounts

    @property
    def counts(self) -> Counter:
        """Return the counts object."""
        return self._counts



"""Enum classes for Pysiaalarm."""
from __future__ import annotations

from enum import Enum, auto
from typing import Any


class AutoName(Enum):
    """Auto name the enums."""

    @staticmethod
    def _generate_next_value_(
        name: str,
        start: Any,
        count: Any,
        last_values: Any,  # pylint: disable=unused-argument
    ) -> str:
        """Return name as value for Enum."""
        return name


class CommunicationsProtocol(AutoName):
    """CommunicationsProtocol enumerator for SIA."""

    TCP = auto()
    UDP = auto()


class ResponseType(AutoName):
    """Class with response types for events."""

    ACK = auto()
    DUH = auto()
    NAK = auto()
    RSP = auto()


class MessageTypes(Enum):
    """Message type enumerator for SIA."""

    SIADCS = "SIA-DCS"
    ADMCID = "ADM-CID"
    NULL = "NULL"
    OH = "OH"



"""This is the class for the actual TCP handler override of the handle method."""
from __future__ import annotations

import asyncio
import logging
from collections.abc import Awaitable, Callable
from typing import Any

from .. import __author__, __copyright__, __license__, __version__
from ..account import SIAAccount
from ..base_server import BaseSIAServer
from ..const import EMPTY_BYTES
from ..event import SIAEvent
from ..utils import Counter

_LOGGER = logging.getLogger(__name__)


class SIAServerTCP(BaseSIAServer):
    """Class for SIA TCP Server Async."""

    def __init__(
        self,
        accounts: dict[str, SIAAccount],
        func: Callable[[SIAEvent], Awaitable[None]],
        counts: Counter,
    ):
        """Create a SIA TCP Server.

        Arguments:
            accounts Dict[str, SIAAccount] -- accounts as dict with account_id as key, SIAAccount object as value.  # pylint: disable=line-too-long
            func Callable[[SIAEvent], None] -- Function called for each valid SIA event, that can be matched to a account.  # pylint: disable=line-too-long
            counts Counter -- counter kept by client to give insights in how many errorous events were discarded of each type.  # pylint: disable=line-too-long
        """
        BaseSIAServer.__init__(self, accounts, counts, async_func=func)

    async def handle_line(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        """Handle line for SIA Events. This supports TCP connections.

        Arguments:
            reader {asyncio.StreamReader} -- StreamReader with new data.
            writer {asyncio.StreamWriter} -- StreamWriter to respond.

        """
        while True and not self.shutdown_flag:  # pragma: no cover  # type: ignore
            try:
                data = await reader.read(1000)
            except ConnectionResetError:
                break
            if data == EMPTY_BYTES or reader.at_eof():
                break
            event = self.parse_and_check_event(data)
            if not event:
                continue
            writer.write(event.create_response())
            await writer.drain()
            await self.async_func_wrap(event)

        writer.close()


class SIAServerUDP(BaseSIAServer, asyncio.DatagramProtocol):
    """Class for SIA UDP Server Async."""

    def __init__(
        self,
        accounts: dict[str, SIAAccount],
        func: Callable[[SIAEvent], Awaitable[None]],
        counts: Counter,
    ):
        """Create a SIA UDP Server.

        Arguments:
            server_address {tuple(string, int)} -- the address the server should listen on.
            accounts {Dict[str, SIAAccount]} -- accounts as dict with account_id as key, SIAAccount object as value.  # pylint: disable=line-too-long
            func {Callable[[SIAEvent], None]} -- Function called for each valid SIA event, that can be matched to a account.  # pylint: disable=line-too-long
            counts {Counter} -- counter kept by client to give insights in how many errorous events were discarded of each type.  # pylint: disable=line-too-long
        """
        BaseSIAServer.__init__(self, accounts, counts, async_func=func)
        self.transport: asyncio.DatagramTransport | None = None

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Connect callback for datagrams."""
        assert isinstance(transport, asyncio.DatagramTransport)
        self.transport = transport

    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        """Receive and process datagrams. This support UDP connections."""
        event = self.parse_and_check_event(data)
        if not event:
            return
        if self.transport is not None:
            self.transport.sendto(event.create_response(), addr)
        asyncio.create_task(self.async_func_wrap(event))

    def connection_lost(self, _: Any) -> None:
        """Close and reset transport when connection lost."""
        if self.transport:
            self.transport.close()



"""This is a class for SIA Events."""
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field, asdict
from datetime import datetime, timedelta, timezone, tzinfo
from typing import Union, Any

from Crypto.Cipher import AES
from Crypto.Cipher._mode_cbc import CbcMode

from .account import SIAAccount
from .const import IV, RSP_XDATA
from .errors import EventFormatError, NoAccountError
from .utils import (
    MAIN_MATCHER,
    OH_MATCHER,
    MessageTypes,
    ResponseType,
    SIACode,
    SIAXData,
    _get_matcher,
    _load_adm_mapping,
    _load_sia_codes,
    _load_xdata,
)

_LOGGER = logging.getLogger(__name__)


@dataclass  # type: ignore
class BaseEvent(ABC):
    """Base class for Events."""

    # From Main Matcher
    full_message: str | None = None
    msg_crc: str | None = None
    length: str | None = None
    encrypted: bool | None = None
    message_type: str | MessageTypes | None = None
    receiver: str = "R0"
    line: str = "L0"
    account: str | None = None
    sequence: str | None = None

    # Content to be parsed
    content: str | None = None
    encrypted_content: str | None = None

    # From (Encrypted) Content
    ti: str | None = None  # pylint: disable=invalid-name
    id: str | None = None  # pylint: disable=invalid-name
    ri: str | None = None  # pylint: disable=invalid-name
    code: str | None = None
    message: str | None = None
    x_data: str | None = None
    timestamp: datetime | str | None = None

    # From ADM-CID
    event_qualifier: str | None = None
    event_type: str | None = None
    partition: str | None = None

    # Parsed fields
    calc_crc: str | None = None
    extended_data: list[SIAXData] | None = None
    sia_account: SIAAccount | None = field(repr=False, default=None)
    sia_code: SIACode | None = field(default=None, repr=False)

    # Parse flags
    _content_parsed: bool = field(init=False, default=False, repr=False)
    _encrypted_content_decrypted: bool = field(init=False, default=False, repr=False)
    _adm_parsed: bool = field(init=False, default=False, repr=False)
    _sia_added: bool = field(init=False, default=False, repr=False)
    _xdata_parsed: bool = field(init=False, default=False, repr=False)

    @property
    def valid_message(self) -> bool:
        """Return True for OH and NAK Events."""
        return True  # pragma: no cover

    @property
    def code_not_found(self) -> bool:
        """Return True if there is no Code."""
        return bool(self.sia_code is None)

    @property
    @abstractmethod
    def response(self) -> ResponseType | None:
        """Abstract method."""

    @property
    def valid_timestamp(self) -> bool:
        """Return True for OH and NAK Events."""
        return True  # pragma: no cover

    @abstractmethod
    def create_response(self) -> bytes:
        """Abstract method."""

    def set_sia_code(self) -> None:
        """Return the SIA Code object, based on the code field."""
        if self.code:  # pragma: no cover
            self.sia_code = _load_sia_codes().get(self.code)  # pylint: disable=E1101
            self._sia_added = True

    def _get_crypter(self) -> CbcMode | None:
        """Give back a encrypter/decrypter."""
        if not self.sia_account:
            return None  # pragma: no cover
        if not self.sia_account.key_b:
            return None  # pragma: no cover
        cypher = AES.new(self.sia_account.key_b, AES.MODE_CBC, IV)
        if isinstance(cypher, CbcMode):
            return cypher
        return None  # pragma: no cover

    @classmethod
    def from_line(
        cls, incoming: str, accounts: dict[str, SIAAccount] | None = None
    ) -> SIAEvent:
        """Create a Event from a line.

        Arguments:
            incoming {str} -- The line to be parsed.
            accounts {List[SIAAccount]} -- accounts to check against, optional

        Raises:
            EventFormatError: If the event is not formatted according to SIA DC09 or ADM-CID.

        """
        line_match = MAIN_MATCHER.match(incoming)
        if not line_match:
            oh_event = OH_MATCHER.match(incoming)
            if oh_event:
                fields = oh_event.groupdict()
                return OHEvent(
                    full_message=incoming,
                    msg_crc="",
                    message_type=MessageTypes.OH,
                    length=str(len(incoming)),
                    encrypted=False,
                    receiver=fields["receiver"],
                    line=fields["line"],
                    account=fields["account"],
                    id=fields["id"],
                )
            raise EventFormatError(
                f"No matches found, event was not a SIA or ADM Spec event, line was: {incoming}"
            )
        main_content = line_match.groupdict()

        encrypted = bool(main_content["encrypted_flag"])
        acc = main_content["account"]
        sia_account = None
        if accounts and acc:
            sia_account = accounts.get(acc, None)

        return SIAEvent(
            full_message=incoming[8:],
            msg_crc=main_content["crc"],
            length=main_content["length"],
            encrypted=encrypted,
            message_type=main_content["message_type"],
            sequence=main_content["sequence"],
            receiver=main_content["receiver"],
            line=main_content["line"],
            account=acc,
            content=main_content["rest"] if not encrypted else None,
            encrypted_content=main_content["rest"] if encrypted else None,
            sia_account=sia_account,
        )

    @staticmethod
    def _get_timestamp(device_timezone: tzinfo | None = None) -> str:
        """Create a timestamp in the right format."""
        if device_timezone:
            return (
                datetime.now()
                .astimezone(device_timezone)
                .strftime("_%H:%M:%S,%m-%d-%Y")
            )
        return datetime.utcnow().strftime("_%H:%M:%S,%m-%d-%Y")

    @staticmethod
    def _crc_calc(msg: str | None) -> str | None:
        """Calculate the CRC of the msg."""
        if msg is None:  # pragma: no cover
            return None
        crc = 0
        for letter in str.encode(msg):
            temp = letter
            for _ in range(0, 8):
                temp ^= crc & 1
                crc >>= 1
                if (temp & 1) != 0:
                    crc ^= 0xA001
                temp >>= 1
        return ("%x" % crc).upper().zfill(4)

    def to_dict(self, **kwargs: Any) -> dict[str, Any]:
        """Create a dict from the dataclass.

        Kwargs are only there for legacy (after no longer using dataclasses_json),
        so remove any other arguments from this function.
        """
        event = deepcopy(self)
        event.sia_account = None
        if event.timestamp is not None and isinstance(event.timestamp, datetime):
            event.timestamp = event.timestamp.isoformat()
        return asdict(event)

    @classmethod
    def from_dict(cls, event: dict[str, Any]) -> BaseEvent:
        """Create a SIA Event from a dict."""
        if "_content_parsed" in event:
            event.pop("_content_parsed")
        if "_encrypted_content_decrypted" in event:
            event.pop("_encrypted_content_decrypted")
        if "_adm_parsed" in event:
            event.pop("_adm_parsed")
        if "_sia_added" in event:
            event.pop("_sia_added")
        if "_xdata_parsed" in event:
            event.pop("_xdata_parsed")
        if "timestamp" in event and event["timestamp"] is not None:
            event["timestamp"] = datetime.fromisoformat(event["timestamp"])
        return cls(**event)



@dataclass
class SIAEvent(BaseEvent):
    """Class for SIAEvents."""

    def __post_init__(self) -> None:
        """Run post init logic."""
        # Convert the message type to the enum
        if isinstance(self.message_type, str):  # pragma: no cover
            self.message_type = MessageTypes(self.message_type)

        # Calculate the CRC of the message.
        if not self.calc_crc:
            self.calc_crc = self._crc_calc(self.full_message)
        # If there is encrypted content and a key, decrypt
        if self.encrypted_content:
            if not self.sia_account or not self.sia_account.encrypted:
                raise NoAccountError("No account present with encrypted message.")
            if not self._encrypted_content_decrypted:
                self.decrypt_content()
        # If there is content (either after decrypting or directly) parse
        if self.content and not self._content_parsed:
            self.parse_content()
        # If it is a ADM-CID message, map the qualifier and type to a code.
        if (
            self.message_type == MessageTypes.ADMCID
            and self.event_qualifier is not None
            and self.event_type is not None
            and not self._adm_parsed
        ):
            self.parse_adm()
        # If there is a code, map it to the full SIA Code spec.
        if self.code and not self._sia_added:
            self.set_sia_code()
        # If there is x_data, parse it.
        if self.x_data and not self._xdata_parsed:
            self.parse_extended_data()  # pragma: no cover

    @property
    def response(self) -> ResponseType | None:
        """Get the responsetype."""
        if not self.valid_message:
            return None
        if self.code_not_found:
            return ResponseType.DUH
        if not self.sia_account:
            return ResponseType.NAK  # pragma: no cover
        if not self.valid_timestamp:
            return ResponseType.NAK
        if self.extended_data is not None:  # pragma: no cover
            if [x for x in self.extended_data if x.identifier in RSP_XDATA]:
                return ResponseType.RSP
        return ResponseType.ACK

    @property
    def sia_string(self) -> str:  # pragma: no cover
        """Create a string with the SIA codes and some other fields."""
        if self.sia_code:
            return f"Code: {self.code}, Type: {self.sia_code.type}, \
            Description: {self.sia_code.description}"
        return f"Code: {self.code}"

    @property
    def valid_length(self) -> bool:
        """Check if the length of the message is the same in the message and supplied.

        Will not throw an error if not correct.
        """
        if self.length is None or self.full_message is None:  # pragma: no cover
            return True
        return int(self.length) == int(
            str(len(self.full_message)), 16
        )  # pragma: no cover

    @property
    def valid_message(self) -> bool:
        """Check the validity of the message by comparing the sent CRC with the calculated CRC."""
        return self.msg_crc == self.calc_crc

    @property
    def valid_timestamp(self) -> bool:
        """Check if the timestamp is within bounds."""
        if not self.sia_account:  # pragma: no cover
            return True
        if self.sia_account.allowed_timeband is None:  # pragma: no cover
            return True
        if self.timestamp and isinstance(self.timestamp, datetime):
            current_time = datetime.now(self.sia_account.device_timezone)
            current_min = current_time - timedelta(
                seconds=self.sia_account.allowed_timeband[0]
            )
            current_plus = current_time + timedelta(
                seconds=self.sia_account.allowed_timeband[1]
            )
            return current_min <= self.timestamp <= current_plus
        return True  # pragma: no cover

    def create_response(self) -> bytes:
        """Create a response message, based on account, event and response type.

        Returns:
            bytes -- Response to send back to sender.

        """
        response_type = self.response
        x_data = None
        if response_type is None:
            return b"\n\r"
        if not self.sia_account:
            return f'"{response_type.value}"'.encode("ascii")
        if (
            self.extended_data
            and [x for x in self.extended_data if x.identifier == "K"] is not None
            and self.sia_account.key is not None
        ):
            x_data = f"[K{self.sia_account.key}]"
        if response_type == ResponseType.NAK:
            res = f'"{response_type.value}"0000{self.receiver}{self.line}A0[]{self._get_timestamp(self.sia_account.device_timezone)}'
        elif not self.encrypted or response_type == ResponseType.DUH:
            res = f'"{response_type.value}"{self.sequence}{self.receiver}{self.line}#{self.account}[]{x_data if x_data else ""}'
        else:
            encrypted_content = self.encrypt_content(
                f']{x_data if x_data else ""}{self._get_timestamp(self.sia_account.device_timezone)}'
            )
            res = f'"*{response_type.value}"{self.sequence}{self.receiver}{self.line}#{self.account}[{encrypted_content}'
        header = ("%04x" % len(res)).upper()
        new_crc = self._crc_calc(res)
        return f"\n{new_crc}{header}{res}\r".encode("ascii")

    def decrypt_content(self) -> None:
        """Decrypt the content, if encrypted account, otherwise pass back the event."""
        if not self.encrypted_content:  # pragma: no cover
            return None
        decr = self._get_crypter()
        if not decr:
            return None  # pragma: no cover
        self.content = decr.decrypt(bytes.fromhex(self.encrypted_content)).decode(
            "ascii", "ignore"
        )
        self._encrypted_content_decrypted = True

    def encrypt_content(self, message: str) -> str | None:
        """Encrypt a string.

        Arguments:
            message {str} -- String to encrypt.

        Returns:
            str -- Encrypted string, if encrypted account.

        """
        encr = self._get_crypter()
        if not encr:
            return None  # pragma: no cover
        fill_size = len(message) + 16 - len(message) % 16
        return encr.encrypt(message.zfill(fill_size).encode("ascii")).hex().upper()

    def parse_adm(self) -> None:
        """Parse the event qualifier and type for ADM messages."""
        if self.event_qualifier and self.event_type:  # pragma: no cover
            sub_map = _load_adm_mapping().get(self.event_type, None)
            if sub_map:
                self.code = sub_map.get(self.event_qualifier, None)
        self._adm_parsed = True

    def parse_content(self) -> None:
        """Set internal content field, parse the content and store the right things."""
        if (
            not self.message_type or self.encrypted is None or not self.content
        ):  # pragma: no cover
            return
        matcher = _get_matcher(self.message_type, self.encrypted)
        matches = matcher.match(self.content)
        if not matches:
            raise EventFormatError(
                f"Parse content: no matches found in {self.content}, using matcher: {matcher}"
            )
        content = matches.groupdict()
        _LOGGER.debug("Content matches: %s", content)
        # Parse specific fields per message type
        if self.message_type == MessageTypes.ADMCID:
            self.partition = content["partition"]
            self.event_qualifier = content["event_qualifier"]
            self.event_type = content["event_type"]
        else:
            self.code = content["code"]
            self.ti = content["ti"]
            self.id = content["id"]
            self.message = content["message"]

        # Parse generic fields
        if not self.account:
            self.account = content["account"]
        self.ri = content["ri"]
        self.x_data = content["xdata"]
        if content["timestamp"]:
            try:
                timestamp = datetime.strptime(content["timestamp"], "%H:%M:%S,%m-%d-%Y")
                self.timestamp = timestamp.replace(tzinfo=timezone.utc)
            except ValueError:
                _LOGGER.warning(
                    "Timestamp could not be parsed as a timestamp: %s",
                    content["timestamp"],
                )
        if self.message_type == MessageTypes.NULL and self.code_not_found:
            self.code = "RP"
            self.ri = "0"
        self._content_parsed = True

    def parse_extended_data(self) -> None:
        """Set extended data."""
        if self.x_data is None:  # pragma: no cover
            return
        x_data_list = self.x_data.split("][")
        self.extended_data = []
        for x_data in x_data_list:  # pragma: no cover
            xdata = _load_xdata().get(x_data[0], None)
            if xdata:
                xdata.value = x_data[1:]
                self.extended_data.append(xdata)
        self._xdata_parsed = True

    def sia_account_from_message(self) -> SIAAccount | None:  # pragma: no cover
        """Return the SIA Account.

        If there is not account added, create one based on the account in the message.
        """
        if self.account is not None:
            return SIAAccount(self.account)
        return None

    def __str__(self) -> str:
        """Return the event as a string."""
        return f"\
Content: {self.content}, \
Zone (ri): {self.ri}, \
Code: {self.code}, \
Message: {self.message if self.message else ''}, \
Account: {self.account}, \
Receiver: {self.receiver}, \
Line: {self.line}, \
Timestamp: {self.timestamp}, \
Length: {self.length}, \
Sequence: {self.sequence}, \
CRC: {self.msg_crc}, \
Calc CRC: {self.calc_crc}, \
Encrypted Content: {self.encrypted_content}, \
Full Message: {self.full_message}."


@dataclass
class OHEvent(SIAEvent):
    """Class for OH events."""

    code: str = "RP"

    def __post_init__(self) -> None:
        """If there is a code, map it to the full SIA Code spec."""
        if isinstance(self.message_type, str):  # pragma: no cover
            self.message_type = MessageTypes(self.message_type)

        if not self._sia_added:  # pragma: no cover
            self.set_sia_code()

    @property
    def response(self) -> ResponseType:
        """Return ACK for OH Events."""
        return ResponseType.ACK  # pragma: no cover

    def create_response(self) -> bytes:
        """Create a response message, based on account, event and response type.

        Returns:
            str -- Response to send back to sender.

        """
        return '"ACK"'.encode("ascii")  # pragma: no cover


@dataclass
class NAKEvent(BaseEvent):
    """Class for NAK Events."""

    def __post_init__(self) -> None:
        """Run post init work."""
        if isinstance(self.message_type, str):  # pragma: no cover
            self.message_type = MessageTypes(self.message_type)

    @property
    def response(self) -> ResponseType:
        """Return NAK for NAK Events."""
        return ResponseType.NAK

    def create_response(self) -> bytes:
        """Create a response message, based on account, event and response type.

        Returns:
            str -- Response to send back to sender.

        """
        res = f'"NAK"0000L0R0A0[]{self._get_timestamp()}'
        header = ("%04x" % len(res)).upper()
        return f"\n{self._crc_calc(res)}{header}{res}\r".encode("ascii")


EventsType = Union[SIAEvent, OHEvent, NAKEvent]



"""This is the class for the actual TCP handler override of the handle method."""
from __future__ import annotations

import logging
from collections.abc import Callable
from socketserver import ThreadingTCPServer, ThreadingUDPServer

from ..account import SIAAccount
from ..base_server import BaseSIAServer
from ..event import SIAEvent
from ..utils import Counter
from .handler import SIATCPHandler, SIAUDPHandler

_LOGGER = logging.getLogger(__name__)


class SIATCPServer(ThreadingTCPServer, BaseSIAServer):
    """Class for a threaded SIA TCP Server."""

    daemon_threads = True
    allow_reuse_address = True

    def __init__(
        self,
        server_address: tuple[str, int],
        accounts: dict[str, SIAAccount],
        func: Callable[[SIAEvent], None],
        counts: Counter,
    ):
        """Create a SIA TCP Server.

        Arguments:
            server_address Tuple[string, int] -- the address the server should listen on.
            accounts Dict[str, SIAAccount] -- accounts as dict with account_id as key, SIAAccount object as value.  # pylint: disable=line-too-long
            func Callable[[SIAEvent], None] -- Function called for each valid SIA event, that can be matched to a account.  # pylint: disable=line-too-long
            counts Counter -- counter kept by client to give insights in how many errorous events were discarded of each type.  # pylint: disable=line-too-long
        """
        ThreadingTCPServer.__init__(self, server_address, SIATCPHandler)
        BaseSIAServer.__init__(self, accounts, counts, func=func)


class SIAUDPServer(ThreadingUDPServer, BaseSIAServer):
    """Class for a threaded SIA UDP Server."""

    daemon_threads = True
    allow_reuse_address = True

    def __init__(
        self,
        server_address: tuple[str, int],
        accounts: dict[str, SIAAccount],
        func: Callable[[SIAEvent], None],
        counts: Counter,
    ):
        """Create a SIA UDP Server.

        Arguments:
            server_address Tuple[string, int] -- the address the server should listen on.
            accounts Dict[str, SIAAccount] -- accounts as dict with account_id as key, SIAAccount object as value.  # pylint: disable=line-too-long
            func Callable[[SIAEvent], None] -- Function called for each valid SIA event, that can be matched to a account.  # pylint: disable=line-too-long
            counts Counter -- counter kept by client to give insights in how many errorous events were discarded of each type.  # pylint: disable=line-too-long
        """
        ThreadingUDPServer.__init__(self, server_address, SIAUDPHandler)
        BaseSIAServer.__init__(self, accounts, counts, func=func)



"""Constants for SIA."""
COUNTER_EVENTS = "events"
COUNTER_VALID = "valid_events"
COUNTER_FORMAT = "format"
COUNTER_CRC = "crc"
COUNTER_CODE = "code"
COUNTER_TIMESTAMP = "timestamp"
COUNTER_ACCOUNT = "account"
COUNTER_USER_CODE = "user_code"

IV = bytes.fromhex("00000000000000000000000000000000")
EMPTY_BYTES = b""

RSP_XDATA = ["K"]



#!/usr/bin/python
"""Run a test client."""
import asyncio
import json
import logging
import sys

from .test_utils import create_line_from_test_case, create_random_line


logging.basicConfig(level=logging.DEBUG)
_LOGGER = logging.getLogger(__name__)


async def async_send_messages(config, test_case, time_between):
    """Send message async."""
    host = config["host"]
    port = config["port"]
    _LOGGER.debug("Opening connection")
    reader, writer = await asyncio.open_connection(host, port)
    if test_case:
        _LOGGER.debug("Number of cases: %s", len(test_case))
        for tc in test_case:
            message = create_line_from_test_case(config, tc)
            _LOGGER.debug("Message: %s", message)
            writer.write(message.encode())
            data = await reader.read(100)
            _LOGGER.debug(f"Received from server: {data.decode()}")
            await asyncio.sleep(time_between)
        writer.close()
        return
    try:
        while True:
            message = create_random_line(config)
            _LOGGER.debug("Message: %s", message)
            writer.write(message.encode())
            data = await reader.read(100)
            _LOGGER.debug(f"Received from server: {data.decode()}")
            await asyncio.sleep(time_between)
    finally:
        writer.close()


if __name__ == "__main__":
    """Run main with a config."""
    _LOGGER.info(sys.argv)
    try:  # sys.argv.index(1)
        file = sys.argv[1]
    except:
        file = "tests//local_config.json"
    with open(file, "r") as f:
        config = json.load(f)
    asyncio.get_event_loop().run_until_complete(async_send_messages(config, None, 3))



"""Counter helper class."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from ..const import (
    COUNTER_ACCOUNT,
    COUNTER_CODE,
    COUNTER_CRC,
    COUNTER_EVENTS,
    COUNTER_FORMAT,
    COUNTER_TIMESTAMP,
    COUNTER_USER_CODE,
    COUNTER_VALID,
)


@dataclass
class Counter:
    """Class for the counter."""

    error_account: int = 0
    error_code: int = 0
    error_crc: int = 0
    error_format: int = 0
    error_timestamp: int = 0
    error_user_function: int = 0
    events: int = 0
    valid_events: int = 0

    def increment_error_account(self) -> None:
        """Increment the error_account count."""
        self.error_account += 1

    def increment_error_code(self) -> None:
        """Increment the error_code count."""
        self.error_code += 1

    def increment_error_crc(self) -> None:
        """Increment the error_crc count."""
        self.error_crc += 1

    def increment_error_format(self) -> None:
        """Increment the error_format count."""
        self.error_format += 1

    def increment_error_timestamp(self) -> None:
        """Increment the error_timestamp count."""
        self.error_timestamp += 1

    def increment_error_user_function(self) -> None:
        """Increment the error_user_function count."""
        self.error_user_function += 1

    def increment_events(self) -> None:
        """Increment the events count."""
        self.events += 1

    def increment_valid_events(self) -> None:
        """Increment the valid_events count."""
        self.valid_events += 1

    def get(self, item: str) -> Optional[int]:
        """Get the right counter."""
        if item == COUNTER_ACCOUNT:
            return self.error_account
        if item == COUNTER_CODE:
            return self.error_code
        if item == COUNTER_CRC:
            return self.error_crc
        if item == COUNTER_FORMAT:
            return self.error_format
        if item == COUNTER_TIMESTAMP:
            return self.error_timestamp
        if item == COUNTER_USER_CODE:
            return self.error_user_function
        if item == COUNTER_VALID:
            return self.valid_events
        if item == COUNTER_EVENTS:
            return self.events
        return None  # pragma: no cover

    def increment(self, item: str) -> None:  # pragma: no cover
        """Increment the right counter."""
        if item == COUNTER_ACCOUNT:
            self.increment_error_account()
            return
        if item == COUNTER_CODE:
            self.increment_error_code()
            return
        if item == COUNTER_CRC:
            self.increment_error_crc()
            return
        if item == COUNTER_FORMAT:
            self.increment_error_format()
            return
        if item == COUNTER_TIMESTAMP:
            self.increment_error_timestamp()
            return
        if item == COUNTER_USER_CODE:
            self.increment_error_user_function()
            return
        if item == COUNTER_VALID:
            self.increment_valid_events()
            return
        if item == COUNTER_EVENTS:
            self.increment_events()



# -*- coding: utf-8 -*-
"""This is a the main class for the SIA Client."""
from __future__ import annotations

import asyncio
from collections.abc import Callable
import logging
from threading import Thread
from types import TracebackType
from typing import Any, Type

from ..account import SIAAccount
from ..base_client import BaseSIAClient
from ..event import SIAEvent
from ..utils import CommunicationsProtocol
from .server import SIATCPServer, SIAUDPServer

_LOGGER = logging.getLogger(__name__)


class SIAClient(Thread, BaseSIAClient):
    """Class for Sync SIA Client."""

    def __init__(
        self,
        host: str,
        port: int,
        accounts: list[SIAAccount],
        function: Callable[[SIAEvent], None],
        protocol: CommunicationsProtocol = CommunicationsProtocol.TCP,
    ):
        """Create the threaded SIA Client object.

        Arguments:
            host {str} -- Host to run the server on, usually would be ""
            port {int} -- The port the server listens to.
            accounts {List[SIAAccount]} -- List of SIA Accounts to add.
            function {Callable[[SIAEvent], None]} -- The function that gets called for each event.
            protocol {CommunicationsProtocol Enum} -- CommunicationsProtocol to use, TCP or UDP.

        """
        if asyncio.iscoroutinefunction(function):
            raise TypeError(
                "Asyncio coroutines as the function are not supported, please use the aio version of the SIAClient for that."  # pylint: disable=line-too-long
            )
        Thread.__init__(self)
        BaseSIAClient.__init__(self, host, port, accounts, protocol)
        self._func: Callable[[SIAEvent], None] = function
        self.sia_server: SIATCPServer | SIAUDPServer = self.get_server()
        self.server_thread: Thread | None = None

    def get_server(self) -> SIATCPServer | SIAUDPServer:
        """Set the sia server to a TCP server."""
        if self.protocol == CommunicationsProtocol.TCP:
            return SIATCPServer(  # type: ignore
                (self._host, self._port), self._accounts, self._func, self._counts
            )
        return SIAUDPServer(  # type: ignore
            (self._host, self._port), self._accounts, self._func, self._counts
        )

    def __enter__(self) -> SIAClient:
        """Start with as context manager."""
        self.start()
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        """End as context manager."""
        self.stop()
        return True

    def start(self, **kwargs: Any) -> None:
        """Start the SIA Handler thread."""
        _LOGGER.debug("Starting SIA.")
        if self.sia_server is not None:  # pragma: no cover
            self.server_thread = Thread(
                target=self.sia_server.serve_forever,
                name="SIAServerThread",
                kwargs=kwargs,
            )
            self.server_thread.daemon = True
            self.server_thread.start()

    def stop(self) -> None:
        """Stop the SIA TCP Handler thread."""
        _LOGGER.debug("Stopping SIA.")
        if self.sia_server is not None:  # pragma: no cover
            self.sia_server.shutdown_flag = True
            self.sia_server.shutdown()
            self.sia_server.server_close()
        if self.server_thread is not None:  # pragma: no cover
            self.server_thread.join()



"""Helper script with all Regexes."""
from __future__ import annotations

import re

from .enums import MessageTypes

# pylint: disable=invalid-name

# sample OH: SR0001L0001    006969XX    [ID00000000]
oh_regex = r"""
^S
(?:R)(?P<receiver>(?<=R)(?:\d{4}))
(?:L)(?P<line>(?<=L)(?:\d{4}))
\s+(?P<account>\w{8})\s+
\[(?P<id>\w+)\]$
"""
OH_MATCHER = re.compile(oh_regex, re.X)

main_regex = r"""
(?P<crc>[A-Fa-f0-9]{4})
(?P<length>[A-Fa-f0-9]{4})\"
(?P<encrypted_flag>[*])?
(?P<message_type>SIA-DCS|ADM-CID|NULL)\"
(?P<sequence>[0-9]{4})
(?P<receiver>R[A-Fa-f0-9]{1,6})?
(?P<line>L[A-Fa-f0-9]{1,6})
[#]?(?P<account>[A-Fa-f0-9]{3,16})?
[\[]
(?P<rest>.*)
"""
MAIN_MATCHER = re.compile(main_regex, re.X)

sia_content_regex = r"""
[#]?(?P<account>[A-Fa-f0-9]{3,16})?
[|]?
[N]?
(?:ti)?(?:(?<=ti)(?P<ti>\d{2}:\d{2}))?\/?
(?:id)?(?:(?<=id)(?P<id>\d*))?\/?
(?:ri)?(?:(?<=ri)(?P<ri>\d*))?\/?
(?P<code>[a-zA-Z]{2})?
(?P<message>.[^\[\]]*)?
[\]]
(?:\[(?:(?<=\[)(?P<xdata>.*)(?=\]))\])?
[_]?
(?P<timestamp>[0-9:,-]*)?$
"""
SIA_CONTENT_MATCHER = re.compile(sia_content_regex, re.X)

encr_sia_content_regex = r"""(?:[^\|\[\]]*)[|]?"""
ENCR_SIA_CONTENT_MATCHER = re.compile(encr_sia_content_regex + sia_content_regex, re.X)

adm_content_regex = r"""
[#]?(?P<account>[A-F0-9]{3,16})?
[|]?
(?P<event_qualifier>\d{1})
(?P<event_type>\d{3})
\s
(?P<partition>\d{2})
\s
(?P<ri>\d{3})
[\]]
(?:\[(?:(?<=\[)(?P<xdata>.[^\[\]]*)(?=\]))\])?
[_]?
(?P<timestamp>[0-9:,-]*)?$
"""
ADM_CONTENT_MATCHER = re.compile(adm_content_regex, re.X)

encr_adm_content_regex = r"""(?:[^\|\[\]]*)[|]?"""
ENCR_ADM_CONTENT_MATCHER = re.compile(encr_adm_content_regex + adm_content_regex, re.X)


def _get_matcher(
    message_type: MessageTypes | str, encrypted: bool = False
) -> re.Pattern:
    """Extract the content using the different regexes."""
    if message_type == MessageTypes.ADMCID:
        return ENCR_ADM_CONTENT_MATCHER if encrypted else ADM_CONTENT_MATCHER
    return ENCR_SIA_CONTENT_MATCHER if encrypted else SIA_CONTENT_MATCHER
